# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function

import os
import re
import logging
import codecs
from tempfile import mktemp
from xml.sax.saxutils import escape as xml_escape

from rdflib import Graph, URIRef, Literal
from bs4 import BeautifulSoup
import requests
import six

from ferenda import TextReader
from ferenda.sources.legal.se.legalref import LegalRef
from ferenda import util
from . import SwedishLegalSource


class MyndFskr(SwedishLegalSource):

    """A abstract base class for fetching and parsing regulations from
various swedish government agencies. These PDF documents often have
a similar structure both graphically and linguistically, enabling us
to parse them in a generalized way. (Downloading them often requires
special-case code, though.)"""
    source_encoding = "utf-8"
    downloaded_suffix = ".pdf"
    alias = 'myndfskr'

    def download(self, basefile=None):
        """Simple default implementation that downloads all PDF files
        from self.start_url that look like regulation document
        numbers."""
        resp = requests.get(self.start_url)
        # regex to search the link url, text or title for something
        # looking like a FS number
        re_fsnr = re.compile('(\d{4})[:/_-](\d+)(|\.\w+)$')
        tree = lxml.html.document_fromstring(resp.text)
        tree.make_links_absolute(url, resolve_base_href=True)
        for element, attribute, link, pos in tree.iterlinks():
            if link[-4:].lower() != ".pdf":
                continue
            done = False
            # print "Examining %s"  % link
            attrs = dict(link.attrs)
            flds = [link.url, link.text]
            if 'title' in attrs:
                flds.append(attrs['title'])
            for fld in flds:
                if re_fsnr.search(fld) and not done:
                    m = re_fsnr.search(fld)
                    # Make sure we end up with "2011:4" rather than
                    # "2011:04"
                    basefile = "%s:%s" % (m.group(1), int(m.group(2)))
                    self.download_single(basefile, usecache, link.absolute_url)
                    done = True

    def canonical_uri(self, basefile):
        # The canonical URI for these documents cannot always be
        # computed from the basefile. Find the primary subject of the
        # distilled RDF graph instead.
        if not os.path.exists(self.store.distilled_path(basefile)):
            return None

        g = Graph()
        g.parse(self.store.distilled_path(basefile))
        subjects = list(g.subject_objects(self.ns['rdf']['type']))

        if subjects:
            return str(subjects[0][0])
        else:
            self.log.warning(
                "No canonical uri in %s" % (self.distilled_path(basefile)))
            # fall back
            return super(MyndFskr, self).canonical_uri(basefile)

    def textreader_from_basefile(self, basefile, encoding):
        infile = self.store.downloaded_path(basefile)
        tmpfile = self.store.path(basefile, 'intermediate', '.pdf')
        outfile = self.store.path(basefile, 'intermediate', '.txt')
        util.copy_if_different(infile, tmpfile)
        util.runcmd("pdftotext %s" % tmpfile, require_success=True)
        util.robust_remove(tmpfile)

        return TextReader(outfile, encoding=encoding, linesep=TextReader.UNIX)

    def rpubl_uri_transform(self, s):
        # Inspired by
        # http://code.activestate.com/recipes/81330-single-pass-multiple-replace/
        table = {'å': 'aa',
                 'ä': 'ae',
                 'ö': 'oe'}
        r = re.compile("|".join(list(table.keys())))
        # return r.sub(lambda f: table[f.string[f.start():f.end()]], s.lower())
        return r.sub(lambda m: table[m.group(0)], s.lower())

    def download_resource_lists(self, resource_url, graph_path):
        hdr = self._addheaders()
        hdr['Accept'] = 'application/rdf+xml'
        resp = requests.get(resource_url, headers=hdr)
        g = Graph()
        g.parse(data=resp.text, format="xml")
        for subj in g.subjects(self.ns['rdf'].type,
                               self.ns['rpubl'].Forfattningssamling):
            resp = requests.get(str(subj), headers=hdr)
            resp.encoding = "utf-8"
            g.parse(data=resp.text, format="xml")
        with open(graph_path, "wb") as fp:
            data = g.serialize(format="xml")
            fp.write(data)

    def parse_from_textreader(self, reader, basefile):
        tracelog = logging.getLogger("%s.tracelog" % self.alias)

        doc = self.make_document(basefile)
        g = doc.meta

        # 1.2: Load known entities and their URIs (we have to add some
        # that are not yet in the official resource lists
        resource_list_file = self.store.path('resourcelist', 'intermediate', '.rdf')
        if not os.path.exists(resource_list_file):
            self.download_resource_lists("http://service.lagrummet.se/var/common",
                                         resource_list_file)
        resources = Graph()
        resources.parse(resource_list_file, format="xml")

        # 1.3: Define regexps for the data we search for.
        fwdtests = {'dct:issn': ['^ISSN (\d+\-\d+)$'],
                    'dct:title': ['((?:Föreskrifter|[\w ]+s (?:föreskrifter|allmänna råd)).*?)\n\n'],
                    'dct:identifier': ['^([A-ZÅÄÖ-]+FS\s\s?\d{4}:\d+)$'],
                    'rpubl:utkomFranTryck': ['Utkom från\strycket\s+den\s(\d+ \w+ \d{4})'],
                    'rpubl:omtryckAv': ['^(Omtryck)$'],
                    'rpubl:genomforDirektiv': ['Celex (3\d{2,4}\w\d{4})'],
                    'rpubl:beslutsdatum': ['(?:har beslutats|beslutade|beslutat) den (\d+ \w+ \d{4})'],
                    'rpubl:beslutadAv': ['\n([A-ZÅÄÖ][\w ]+?)\d? (?:meddelar|lämnar|föreskriver)',
                                         '\s(?:meddelar|föreskriver) ([A-ZÅÄÖ][\w ]+?)\d?\s'],
                    'rpubl:bemyndigande': [' ?(?:meddelar|föreskriver|Föreskrifterna meddelas|Föreskrifterna upphävs)\d?,? (?:följande |)med stöd av\s(.*?) ?(?:att|efter\ssamråd|dels|följande|i fråga om|och lämnar allmänna råd|och beslutar följande allmänna råd|\.\n)',
                                           '^Med stöd av (.*)\s(?:meddelar|föreskriver)']
                    }

        # 2: Find metadata properties

        # 2.1 Find some of the properties on the first page (or the
        # 2nd, or 3rd... continue past TOC pages, cover pages etc
        # until the "real" first page is found) NB: FFFS 2007:1 has
        # ten (10) TOC pages!
        pagecount = 0
        for page in reader.getiterator(reader.readpage):
            # replace single newlines with spaces, but keep double
            # newlines
            # page = "\n\n".join([util.normalize_space(x) for x in page.split("\n\n")])
            pagecount += 1
            props = {}
            for (prop, tests) in list(fwdtests.items()):
                if prop in props:
                    continue
                for test in tests:
                    m = re.search(
                        test, page, re.MULTILINE | re.DOTALL | re.UNICODE)
                    if m:
                        props[prop] = util.normalize_space(m.group(1))
            # Single required propery. If we find this, we're done
            if 'rpubl:beslutsdatum' in props:
                break
            self.log.warning("%s: Couldn't find required props on page %s" %
                             (basefile, pagecount))

        # 2.2 Find some of the properties on the last 'real' page (not
        # counting appendicies)
        reader.seek(0)
        pagesrev = reversed(list(reader.getiterator(reader.readpage)))
        # The language used to expres these two properties differ
        # quite a lot, more than what is reasonable to express in a
        # single regex. We therefore define a set of possible
        # expressions and try them in turn.
        revtests = {'rpubl:ikrafttradandedatum':
                    ['(?:Denna författning|Dessa föreskrifter|Dessa allmänna råd|Dessa föreskrifter och allmänna råd)\d* träder i ?kraft den (\d+ \w+ \d{4})',
                     'Dessa föreskrifter träder i kraft, (?:.*), i övrigt den (\d+ \w+ \d{4})',
                     'ska(?:ll|)\supphöra att gälla (?:den |)(\d+ \w+ \d{4}|denna dag|vid utgången av \w+ \d{4})',
                     'träder i kraft den dag då författningen enligt uppgift på den (utkom från trycket)'],
                    'rpubl:upphaver':
                    ['träder i kraft den (?:\d+ \w+ \d{4}), då(.*)ska upphöra att gälla',
                     'ska(?:ll|)\supphöra att gälla vid utgången av \w+ \d{4}, nämligen(.*?)\n\n',
                     'att (.*) skall upphöra att gälla (denna dag|vid utgången av \w+ \d{4})']
                    }

        cnt = 0
        for page in pagesrev:
            cnt += 1
            # Normalize the whitespace in each paragraph so that a
            # linebreak in the middle of the natural language
            # expression doesn't break our regexes.
            page = "\n\n".join(
                [util.normalize_space(x) for x in page.split("\n\n")])

            for (prop, tests) in list(revtests.items()):
                if prop in props:
                    continue
                for test in tests:
                    # Not re.DOTALL -- we've normalized whitespace and
                    # don't want to match across paragraphs
                    m = re.search(test, page, re.MULTILINE | re.UNICODE)
                    if m:
                        props[prop] = util.normalize_space(m.group(1))
                        # print u"%s: '%s' resulted in match '%s' at page %s from end" %
                        # (prop,test,props[prop], cnt)

            # Single required propery. If we find this, we're done
            if 'rpubl:ikrafttradandedatum' in props:
                break

        # 3: Clean up data - converting strings to Literals or
        # URIRefs, find legal references, etc
        if 'dct:identifier' in props:
            (publication, year, ordinal) = re.split('[ :]',
                                                    props['dct:identifier'])
            # FIXME: Read resources graph instead
            fs = resources.value(predicate=self.ns['skos'].altLabel,
                                 object=Literal(publication, lang='sv'))
            props['rpubl:forfattningssamling'] = fs
            publ = resources.value(subject=fs,
                                   predicate=self.ns['dct'].publisher)
            props['dct:publisher'] = publ

            props['rpubl:arsutgava'] = Literal(
                year)  # conversion to int, date not needed
            props['rpubl:lopnummer'] = Literal(ordinal)
            props['dct:identifier'] = Literal(props['dct:identifier'])

            # Now we can mint the uri (should be done through LegalURI)
            uri = ("http://rinfo.lagrummet.se/publ/%s/%s:%s" %
                   (props['rpubl:forfattningssamling'].split('/')[-1],
                    props['rpubl:arsutgava'],
                    props['rpubl:lopnummer']))
            self.log.debug("URI: %s" % uri)
        else:
            self.log.error(
                "Couldn't find dct:identifier, cannot create URI, giving up")
            return None

        tracelog.info("Cleaning rpubl:beslutadAv")
        if 'rpubl:beslutadAv' in props:
            agency = resources.value(predicate=self.ns['foaf'].name,
                                     object=Literal(props['rpubl:beslutadAv'],
                                                    lang="sv"))
            if agency:
                props['rpubl:beslutadAv'] = agency
            else:
                self.log.warning(
                    "Cannot find URI for rpubl:beslutadAv value %r" % props['rpubl:beslutadAv'])
                del props['rpubl:beslutadAv']

        tracelog.info("Cleaning dct:issn")
        if 'dct:issn' in props:
            props['dct:issn'] = Literal(props['dct:issn'])

        tracelog.info("Cleaning dct:title")

        # common false positive
        if 'dct:title' in props and 'denna f\xf6rfattning har beslutats den' in props['dct:title']:
            del props['dct:title']

        if 'dct:title' in props:
            tracelog.info("Inspecting dct:title %r" % props['dct:title'])
            # sometimes the title isn't separated with two newlines from the rest of the text
            if "\nbeslutade den " in props['dct:title']:
                props['dct:title'] = props[
                    'dct:title'].split("\nbeslutade den ")[0]
            props['dct:title'] = Literal(
                util.normalize_space(props['dct:title']), lang="sv")

            if re.search('^(Föreskrifter|[\w ]+s föreskrifter) om ändring i ', props['dct:title'], re.UNICODE):
                tracelog.info("Finding rpubl:andrar in dct:title")
                orig = re.search(
                    '([A-ZÅÄÖ-]+FS \d{4}:\d+)', props['dct:title']).group(0)
                (publication, year, ordinal) = re.split('[ :]', orig)
                origuri = "http://rinfo.lagrummet.se/publ/%s/%s:%s" % (self.rpubl_uri_transform(publication),
                                                                       year, ordinal)
                props['rpubl:andrar'] = URIRef(origuri)
                if 'rpubl:omtryckAv' in props:
                    props['rpubl:omtryckAv'] = URIRef(origuri)
            if (re.search('^(Föreskrifter|[\w ]+s föreskrifter) om upphävande av', props['dct:title'], re.UNICODE)
                    and not 'rpubl:upphaver' in props):
                tracelog.info("Finding rpubl:upphaver in dct:title")
                props['rpubl:upphaver'] = six.text_type(
                    props['dct:title'])  # cleaned below

        tracelog.info("Cleaning date properties")
        for prop in ('rpubl:utkomFranTryck', 'rpubl:beslutsdatum', 'rpubl:ikrafttradandedatum'):
            if prop in props:
                if (props[prop] == 'denna dag' and
                        prop == 'rpubl:ikrafttradandedatum'):
                    props[prop] = props['rpubl:beslutsdatum']
                elif (props[prop] == 'utkom från trycket' and
                      prop == 'rpubl:ikrafttradandedatum'):
                    props[prop] = props['rpubl:utkomFranTryck']
                else:
                    props[prop] = Literal(
                        self.parse_swedish_date(props[prop].lower()))

        tracelog.info("Cleaning rpubl:genomforDirektiv")
        if 'rpubl:genomforDirektiv' in props:
            props['rpubl:genomforDirektiv'] = URIRef("http://rinfo.lagrummet.se/ext/eur-lex/%s" %
                                                     props['rpubl:genomforDirektiv'])

        tracelog.info("Cleaning rpubl:bemyndigande")
        has_bemyndiganden = False

        if 'rpubl:bemyndigande' in props:
            # SimpleParse can't handle unicode endash sign, transform
            # into regular ascii hyphen
            props['rpubl:bemyndigande'] = props[
                'rpubl:bemyndigande'].replace('\u2013', '-')
            parser = LegalRef(LegalRef.LAGRUM)
            result = parser.parse(props['rpubl:bemyndigande'])
            bemyndigande_uris = [x.uri for x in result if hasattr(x, 'uri')]

            # some of these uris need to be filtered away due to
            # over-matching by parser.parse
            filtered_bemyndigande_uris = []
            for bem_uri in bemyndigande_uris:
                keep = True
                for compare in bemyndigande_uris:
                    if (len(compare) > len(bem_uri) and
                            compare.startswith(bem_uri)):
                        keep = False
                if keep:
                    filtered_bemyndigande_uris.append(bem_uri)

            for bem_uri in filtered_bemyndigande_uris:
                g.add((URIRef(
                    uri), self.ns['rpubl']['bemyndigande'], URIRef(bem_uri)))
                has_bemyndiganden = True
            del props['rpubl:bemyndigande']

        tracelog.info("Cleaning rpubl:upphaver")
        if 'rpubl:upphaver' in props:
            for upph in re.findall('([A-ZÅÄÖ-]+FS \d{4}:\d+)', util.normalize_space(props['rpubl:upphaver'])):
                (publication, year, ordinal) = re.split('[ :]', upph)
                upphuri = "http://rinfo.lagrummet.se/publ/%s/%s:%s" % (publication.lower(),
                                                                       year, ordinal)
                g.add((URIRef(
                    uri), self.ns['rpubl']['upphaver'], URIRef(upphuri)))
            del props['rpubl:upphaver']

        tracelog.info("Deciding rdf:type")
        if ('dct:title' in props and
            "allmänna råd" in props['dct:title'] and
                not "föreskrifter" in props['dct:title']):
            props['rdf:type'] = self.ns['rpubl']['AllmannaRad']
        else:
            props['rdf:type'] = self.ns['rpubl']['Myndighetsforeskrift']

        # 3.5: Check to see that we have all properties that we expect
        # (should maybe be done elsewhere later?)
        tracelog.info("Checking required properties")
        for prop in ('dct:identifier', 'dct:title', 'rpubl:arsutgava',
                     'dct:publisher', 'rpubl:beslutadAv', 'rpubl:beslutsdatum',
                     'rpubl:forfattningssamling', 'rpubl:ikrafttradandedatum',
                     'rpubl:lopnummer', 'rpubl:utkomFranTryck'):
            if not prop in props:
                self.log.warning("%s: Failed to find %s" % (basefile, prop))

        tracelog.info("Checking rpubl:bemyndigande")
        if props['rdf:type'] == self.ns['rpubl']['Myndighetsforeskrift']:
            if not has_bemyndiganden:
                self.log.warning(
                    "%s: Failed to find rpubl:bemyndigande" % (basefile))

        # 4: Add the cleaned data to a RDFLib Graph
        # (maybe we should do that as early as possible?)
        tracelog.info("Adding items to rdflib.Graph")
        for (prop, value) in list(props.items()):
            (prefix, term) = prop.split(":", 1)
            p = self.ns[prefix][term]
            if not (isinstance(value, URIRef) or isinstance(value, Literal)):
                self.log.warning("%s: %s is a %s, not a URIRef or Literal" %
                                 (basefile, prop, type(value)))
            g.add((URIRef(uri), p, value))

        # 5: Create data for the body, removing various control characters
        # TODO: Use pdftohtml to create a nice viewable HTML
        # version instead of this plaintext stuff
        reader.seek(0)
        body = []

        # A fairly involved way of filtering out all control
        # characters from a string
        import unicodedata
        if six.PY3:
            all_chars = (chr(i) for i in range(0x10000))
        else:
            all_chars = (unichr(i) for i in range(0x10000))
        control_chars = ''.join(
            c for c in all_chars if unicodedata.category(c) == 'Cc')
        # tab and newline are technically Control characters in
        # unicode, but we want to keep them.
        control_chars = control_chars.replace("\t", "").replace("\n", "")

        control_char_re = re.compile('[%s]' % re.escape(control_chars))
        for page in reader.getiterator(reader.readpage):
            text = xml_escape(control_char_re.sub('', page))
            body.append("<pre>%s</pre>\n\n" % text)

        # 5: Done!
        #
        doc.body = body
        doc.uri = uri
        return doc

    def tabs(cls, primary=False):
        return [['Myndighetsföreskrifter', '/myndfskr/']]


class SJVFS(MyndFskr):
    alias = "sjvfs"
    start_url = "http://www.jordbruksverket.se/forfattningar/forfattningssamling.4.5aec661121e2613852800012537.html"

    def download(self, basefile=None):
        soup = BeautifulSoup(requests.get(self.start_url).text)
        main = soup.find("ul", "c112")
        extra = []
        for a in list(main.findAll("a")):
            url = urllib.parse.urljoin(self.start_url, a['href'])
            self.log.info("Fetching %s %s" % (a.text, url))
            extra.extend(self.download_indexpage(url, usecache=usecache))

        extra2 = []
        for url in list(set(extra)):
            self.log.info("Extrafetching %s" % (url))
            extra2.extend(self.download_indexpage(url, usecache=usecache))

        for url in list(set(extra2)):
            self.log.info("Extra2fetching %s" % (url))
            self.download_indexpage(url, usecache=usecache)

    def download_indexpage(self, url):

        subsoup = BeautifulSoup(requests.get(url).text)
        submain = subsoup.find("div", "pagecontent")
        extrapages = []
        for a in submain.findAll("a"):
            if a['href'].endswith(".pdf") or a['href'].endswith(".PDF"):
                if re.search('\d{4}:\d+', a.text):
                    m = re.search('(\w+FS|) ?(\d{4}:\d+)', a.text)
                    fs = m.group(1).lower()
                    fsnr = m.group(2)
                    if not fs:
                        fs = "sjvfs"
                    basefile = "%s/%s" % (fs, fsnr)
                    suburl = urllib.parse.unquote(
                        urllib.parse.urljoin(url, a['href'])).encode('utf-8')
                    self.download_single(
                        basefile, usecache=usecache, url=suburl)
                elif a.text == "Besult":
                    basefile = a.findParent(
                        "td").findPreviousSibling("td").find("a").text
                    self.log.debug(
                        "Will download beslut to %s (later)" % basefile)
                elif a.text == "Bilaga":
                    basefile = a.findParent(
                        "td").findPreviousSibling("td").find("a").text
                    self.log.debug(
                        "Will download bilaga to %s (later)" % basefile)
                elif a.text == "Rättelseblad":
                    basefile = a.findParent(
                        "td").findPreviousSibling("td").find("a").text
                    self.log.debug(
                        "Will download rättelseblad to %s (later)" % basefile)
                else:
                    self.log.debug("I don't know what to do with %s" % a.text)
            else:
                suburl = urljoin(url, a['href'])
                extrapages.append(suburl)
        return extrapages


class DVFS(MyndFskr):
    alias = "dvfs"


class FFFS(MyndFskr):
    alias = "fffs"
    start_url = "http://www.fi.se/Regler/FIs-forfattningar/Forteckning-FFFS/"
    document_url = "http://www.fi.se/Regler/FIs-forfattningar/Samtliga-forfattningar/%s/"

    def download(self, basefile=None):
        soup = BeautifulSoup(requests.get(self.start_url).text)
        main = soup.find(id="mainarea")
        docs = []
        for numberlabel in main.findAll(text='NUMMER'):
            numberdiv = numberlabel.findParent('div').parent

            typediv = numberdiv.findNextSibling()
            if typediv.find('div', 'FFFSListAreaLeft').get_text(strip=True) != "TYP":
                self.log.error("Expected TYP in div, found %s" %
                               typediv.get_text(strip=True))
                continue

            titlediv = typediv.findNextSibling()
            if titlediv.find('div', 'FFFSListAreaLeft').get_text(strip=True) != "RUBRIK":
                self.log.error("Expected RUBRIK in div, found %s" %
                               titlediv.get_text(strip=True))
                continue

            number = numberdiv.find('div', 'FFFSListAreaRight').get_text(strip=True)
            tmpfile = mktemp()
            snippetfile = self.store.downloaded_path(
                number).replace(".pdf", ".snippet.html")
            fp = codecs.open(tmpfile, "w", encoding="utf-8")
            fp.write(str(numberdiv))
            fp.write(str(typediv))
            fp.write(str(titlediv))
            fp.close()
            util.replace_if_different(tmpfile, snippetfile)

            self.download_single(number, usecache)

    def download_single(self, basefile, usecache=False):
        self.log.debug("%s: download_single..." % basefile)
        pdffile = self.store.downloaded_path(basefile)
        existed = os.path.exists(pdffile)
        if usecache and existed:
            self.log.debug("%s: already exists, not downloading" % basefile)
            return
        snippetfile = pdffile.replace(".pdf", ".snippet.html")
        descriptionfile = pdffile.replace(".pdf", ".html")

        soup = BeautifulSoup(open(snippetfile))
        href = soup.find(text="RUBRIK").findParent(
            "div").findPreviousSibling().find('a')['href']
        url = urljoin("http://www.fi.se/Regler/FIs-forfattningar/Forteckning-FFFS/", href)
        if href.endswith(".pdf"):
            if self.download_if_needed(url, pdffile):
                if existed:
                    self.log.info("%s: downloaded new version from %s" %
                                  (basefile, url))
                else:
                    self.log.info("%s: downloaded from %s" % (basefile, url))

        elif "/Samtliga-forfattningar/" in href:
            self.log.debug("%s: Separate page" % basefile)
            self.download_if_needed(url, descriptionfile)
            soup = BeautifulSoup(open(descriptionfile))
            for link in soup.find("div", id="mainarea").findAll("a"):
                suburl = urljoin(url, link['href']).replace(" ", "%20")
                if link.text == 'Grundförfattning':
                    if self.download_if_needed(suburl, pdffile):
                        self.log.info("%s: downloaded main PDF" % basefile)

                elif link.text == 'Konsoliderad version':
                    conspdffile = pdffile.replace(".pdf", "_k.pdf")
                    if self.download_if_needed(suburl, conspdffile):
                        self.log.info(
                            "%s: downloaded consolidated PDF" % basefile)

                elif link.text == 'Ändringsförfattning':
                    self.log.info("Skipping change regulation")
                elif link['href'].endswith(".pdf"):
                    filename = link['href'].split("/")[-1]
                    otherpdffile = pdffile.replace(".pdf", "-" + filename)
                    if self.download_if_needed(suburl, otherpdffile):
                        self.log.info("%s: downloaded '%s' to %s" %
                                      (basefile, link.text, otherpdffile))

        else:
            self.log.warning("%s: No idea!" % basefile)


class ELSAKFS(MyndFskr):
    alias = "elsakfs"  # real name is ELSÄK-FS, but avoid swedchars, uppercase and dashes
    uri_slug = "elsaek-fs"  # for use in

    start_url = "http://www.elsakerhetsverket.se/sv/Lag-och-ratt/Foreskrifter/Elsakerhetsverkets-foreskrifter-listade-i-nummerordning/"


class NFS(MyndFskr):
    alias = "nfs"

    start_url = "http://www.naturvardsverket.se/sv/Start/Lagar-och-styrning/Foreskrifter-och-allmanna-rad/Foreskrifter/"


class STAFS(MyndFskr):
    alias = "stafs"
    re_identifier = re.compile('STAFS (\d{4})[:/_-](\d+)')

    start_url = "http://www.swedac.se/sv/Det-handlar-om-fortroende/Lagar-och-regler/Alla-foreskrifter-i-nummerordning/"

    def download(self, basefile=None):
        soup = BeautifulSoup(requests.get(self.start_url).text)
        for link in list(soup.find_all("a", href=re.compile('/STAFS/'))):
            basefile = re.search('\d{4}:\d+', link.text).group(0)
            self.download_single(basefile, urljoin(self.start_url, link['href']))

    def download_single(self, basefile, url):
        self.log.info("%s: %s" % (basefile, url))
        consolidated_link = None
        newest = None
        soup = BeautifulSoup(requests.get(url).text)
        for link in soup.find_all("a", text=self.re_identifier):
            self.log.info("   %s: %s %s" % (basefile, link.text, link.url))
            if "konso" in link.text:
                consolidated_link = link
            else:
                m = self.re_identifier.search(link.text)
                assert m
                if link.url.endswith(".pdf"):
                    basefile = m.group(1) + ":" + m.group(2)
                    filename = self.store.downloaded_path(basefile)
                    self.log.info("        Downloading to %s" % filename)
                    self.download_if_needed(link.absolute_url, filename)
                    if basefile > newest:
                        self.log.debug(
                            "%s larger than %s" % (basefile, newest))
                        consolidated_basefile = basefile + \
                            "/konsoliderad/" + basefile
                        newest = basefile
                    else:
                        self.log.debug(
                            "%s not larger than %s" % (basefile, newest))
                else:
                    # not pdf - link to yet another pg
                    subsoup = BeautifulSoup(requests.get(link).text)
                    for sublink in soup.find_all("a", text=self.re_identifier):
                        self.log.info("   Sub %s: %s %s" %
                                      (basefile, sublink.text, sublink['href']))
                        m = self.re_identifier.search(sublink.text)
                        assert m
                        if sublink.url.endswith(".pdf"):
                            subbasefile = m.group(1) + ":" + m.group(2)
                            self.download_if_needed(urljoin(link, sublink['href'], subbasefile))

        if consolidated_link:
            filename = self.store.downloaded_path(consolidated_basefile)
            self.log.info("        Downloading consd to %s" % filename)
            self.download_if_needed(
                consolidated_link.absolute_url, consolidated_basefile, filename=filename)


class SKVFS(MyndFskr):
    alias = "skvfs"
    source_encoding = "utf-8"
    downloaded_suffix = ".pdf"

    # start_url = "http://www.skatteverket.se/rattsinformation/foreskrifter/tidigarear.4.1cf57160116817b976680001670.html"
    # This url contains slightly more (older) links (and a different layout)?
    start_url = "http://www.skatteverket.se/rattsinformation/lagrummet/foreskriftergallande/aldrear.4.19b9f599116a9e8ef3680003547.html"

    # also consolidated versions
    # http://www.skatteverket.se/rattsinformation/lagrummet/foreskrifterkonsoliderade/aldrear.4.19b9f599116a9e8ef3680004242.html

    # URL's are highly unpredictable. We must find the URL for every
    # resource we want to download, we cannot transform the resource
    # id into a URL
    def download(self, basefile=None):
        self.log.info("Starting at %s" % self.start_url)
        years = {}
        soup = BeautifulSoup(requests.get(self.start_url).text)
        for link in sorted(list(soup.find_all("a", text=re.compile('^\d{4}$'))),
                           key=attrgetter('text')):
            year = int(link.text)
            # Documents for the years 1985-2003 are all on one page
            # (with links leading to different anchors). To avoid
            # re-downloading stuff when usecache=False, make sure we
            # haven't seen this url (sans fragment) before
            url = link.absolute_url.split("#")[0]
            if year not in years and url not in list(years.values()):
                self.download_year(year, url)
                years[year] = url

    # just download the most recent year
    def download_new(self):
        self.log.info("Starting at %s" % self.start_url)
        soup = BeautifulSoup(requests.get(self.start_url).text)
        link = sorted(list(soup.find_all("a", text=re.compile('^\d{4}$'))),
                      key=attrgetter('text'), reverse=True)[0]
        self.download_year(int(link.text), link.absolute_url, usecache=True)

    def download_year(self, year, url):
        self.log.info("Downloading year %s from %s" % (year, url))
        soup = BeautifulSoup(requests.get(self.start_url).text)
        for link in soup.find_all("a", text=re.compile('FS \d+:\d+')):
            if "bilaga" in link.text:
                self.log.warning("Skipping attachment in %s" % link.text)
                continue

            # sanitize trailing junk
            linktext = re.match("\w+FS \d+:\d+", link.text).group(0)
            # something like skvfs/2010/23 or rsfs/1996/9
            basefile = linktext.strip(
            ).lower().replace(" ", "/").replace(":", "/")
            self.download_single(
                basefile, link.absolute_url)

    def download_single(self, basefile, url):
        self.log.info("Downloading %s from %s" % (basefile, url))
        self.document_url = url + "#%s"
        html_downloaded = super(
            SKVFS, self).download_single(basefile)
        year = int(basefile.split("/")[1])
        if year >= 2007:  # download pdf as well
            filename = self.store.downloaded_path(basefile)
            pdffilename = os.path.splitext(filename)[0] + ".pdf"
            if not os.path.exists(pdffilename):
                soup = self.soup_from_basefile(basefile)
                pdflink = soup.find(href=re.compile('\.pdf$'))
                if not pdflink:
                    self.log.debug("No PDF file could be found")
                    return html_downloaded
                pdftext = pdflink.get_text(strip=True)
                pdfurl = urljoin(url, pdflink['href'])
                self.log.debug("Found %s at %s" % (pdftext, pdfurl))
                pdf_downloaded = self.download_if_needed(pdfurl, pdffilename)
                return html_downloaded and pdf_downloaded
            else:
                return False
        else:
            return html_downloaded



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sys, os
from ferenda.compat import unittest
if os.getcwd() not in sys.path: sys.path.insert(0,os.getcwd())

from rdflib import Graph

from ferenda.elements import Body
# SUT
from ferenda import Document

class Main(unittest.TestCase):
    def test_create(self):
        doc = Document(uri="http://example.org/",
                       lang="en",
                       basefile="1")
        self.assertEqual(doc.uri, "http://example.org/")
        self.assertEqual(doc.lang, "en")
        self.assertEqual(doc.basefile, "1")

    def test_create_meta(self):
        g = Graph()
        doc = Document(meta=g)
        self.assertIs(g, doc.meta)

    def test_create_body(self):
        b = Body()
        doc = Document(body=b)
        self.assertIs(b, doc.body)
        
    



# -*- coding: utf-8 -*-
"""A small set of generic functions to convert (dicts or dict-like
objects) to URIs. They are usually matched with a corresponding
citationpattern like the ones found in
:py:mod:`ferenda.citationpatterns`. See :doc:`../citationparsing` for
examples.

"""
from __future__ import unicode_literals
from six.moves.urllib_parse import quote


def generic(d):
    """Converts any dict into a URL. The domain (netlog) is always
example.org, and all keys/values of the dict is turned into a
querystring.

    >>> generic({'foo':'1', 'bar':'2'})
    "http://example.org/?foo=1&bar=2"

    """

    querystring = "&".join([quote(k) + "=" + quote(v) for (k, v) in d.items()])
    return "http://example.org/?%s" % querystring


def url(d):
    """Converts a dict with keys ``scheme``, ``netloc``, ``path`` (and
optionally query and/or fragment) into the corresponding URL.

    >>> url({'scheme':'https', 'netloc':'example.org', 'path':'test'})
    "https://example.org/test

    """

    if ('fragment' not in d and 'query' not in d):
        return "%(scheme)s://%(netloc)s%(path)s" % d
    elif 'fragment' not in d:
        return "%(scheme)s://%(netloc)s%(path)s?%(query)s" % d
    elif 'query' not in d:
        return "%(scheme)s://%(netloc)s%(path)s#%(fragment)s" % d
    else:
        return "%(scheme)s://%(netloc)s%(path)s?%(query)s#%(fragment)s" % d


def eulaw(d):
    """Converts a dict with keys like LegalactType, Directive, ArticleId
(produced by :py:data:`ferenda.citationpatterns.eulaw`) into a
CELEX-based URI.

    .. note::

       This is not yet implemented.

    """

    raise NotImplementedError()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sys, os
if os.getcwd() not in sys.path: sys.path.insert(0,os.getcwd())
import pkg_resources
pkg_resources.resource_listdir('ferenda','res')

from pyparsing import Word,nums

from ferenda.compat import unittest

from ferenda.citationparser import CitationParser
from ferenda.uriformatter import URIFormatter
from ferenda.elements import Body, Heading, Paragraph, Footnote, LinkSubject, serialize
import ferenda.uriformats
import ferenda.citationpatterns

class Main(unittest.TestCase):

    def test_parse_recursive(self):
        doc_citation = ("Doc" + Word(nums).setResultsName("ordinal") 
                        + "/" + 
                        Word(nums,exact=4).setResultsName("year")).setResultsName("DocRef")

        def doc_uri_formatter(parts):
            return "http://example.org/docs/%(year)s/%(ordinal)s/" % parts


        doc = Body([Heading(["About Doc 43/2012 and it's interpretation"]),
                    Paragraph(["According to Doc 43/2012",
                               Footnote(["Available at http://example.org/xyz"]),
                               " the bizbaz should be frobnicated"])
                    ])

        result = Body([Heading(["About ",
                                LinkSubject("Doc 43/2012", predicate="dct:references",
                                           uri="http://example.org/docs/2012/43/"),
                                " and it's interpretation"]),
                       Paragraph(["According to ",
                                  LinkSubject("Doc 43/2012", predicate="dct:references",
                                              uri="http://example.org/docs/2012/43/"),
                                  Footnote(["Available at ",
                                            LinkSubject("http://example.org/xyz", 
                                                        predicate="dct:references",
                                                        uri="http://example.org/xyz")
                                            ]),
                                  " the bizbaz should be frobnicated"])
                       ])
        
        cp = CitationParser(ferenda.citationpatterns.url, doc_citation)
        cp.set_formatter(URIFormatter(("url", ferenda.uriformats.url),
                                      ("DocRef", doc_uri_formatter)))
        doc = cp.parse_recursive(doc)
        self.maxDiff = 4096
        self.assertEqual(serialize(doc),serialize(result))

import doctest
from ferenda import citationparser
def load_tests(loader,tests,ignore):
    tests.addTests(doctest.DocTestSuite(citationparser))
    return tests



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from contextlib import contextmanager
import shutil
import os
from tempfile import NamedTemporaryFile
import filecmp

import six
from six.moves.urllib_parse import quote, unquote

from ferenda import util
from ferenda import errors


class DocumentStore(object):

    """
    Unifies handling of reading and writing of various data files
    during the ``download``, ``parse`` and ``generate`` stages.
    
    :param datadir: The root directory (including docrepo path
                    segment) where files are stored.
    :type datadir: str
    :param downloaded_suffix: File suffix for the main source document
                              format. Determines the suffix of
                              downloaded files.
    :type downloaded_suffix: str
    :param storage_policy: Some repositories have documents in several
                           formats, documents split amongst several
                           files or embedded resources. If
                           ``storage_policy`` is set to ``dir``, then
                           each document gets its own directory (the
                           default filename being ``index`` +suffix),
                           otherwise each doc gets stored as a file in
                           a directory with other files.  Affects
                           :py:meth:`~ferenda.DocumentStore.path`
                           (and therefore all other ``*_path``
                           methods)
    :type storage_policy: str
    """

    def __init__(self, datadir, downloaded_suffix=".html", storage_policy="file"):
        self.datadir = datadir  # docrepo.datadir + docrepo.alias
        self.downloaded_suffix = downloaded_suffix
        self.storage_policy = storage_policy

    @contextmanager
    def _open(self, filename, mode):
        if "w" in mode:
            fp = NamedTemporaryFile(mode, delete=False)
            fp.realname = filename
            try:
                yield fp
            finally:
                tempname = fp.name
                fp.close()
                if not os.path.exists(filename) or not filecmp.cmp(tempname, filename):
                    util.ensure_dir(filename)
                    shutil.move(tempname, filename)
                else:
                    os.unlink(tempname)
        else:
            if "a" in mode and not os.path.exists(filename):
                util.ensure_dir(filename)

            fp = open(filename, mode)
            yield fp

    def path(self, basefile, maindir, suffix, version=None, attachment=None, storage_policy=None):
        """Calculate a full filesystem path for the given parameters.

        :param basefile: The basefile of the resource we're calculating a filename for
        :type  basefile: str
        :param maindir: The stage of processing, e.g. ``downloaded`` or ``parsed``
        :type  maindir: str
        :param suffix: Appropriate file suffix, e.g. ``.txt`` or ``.pdf``
        :param version: Optional. The archived version id
        :type  version: str
        :param attachment: Optional. Any associated file needed by the main file. 
        :type  attachment: str
        :param storage_policy: Optional. Used to override `storage_policy` if needed
        :type  attachment: str

        .. note::

           This is a generic method with many parameters. In order to
           keep your code tidy and and loosely coupled to the actual
           storage policy, you should use methods like
           :meth:`~ferenda.DocumentStore.downloaded_path` or :meth:`~ferenda.DocumentStore.parsed_path` when
           possible.
        
        Example:

        >>> d = DocumentStore(datadir="/tmp/base")
        >>> realsep = os.sep
        >>> os.sep = "/"
        >>> d.path('123/a', 'parsed', '.xhtml') == '/tmp/base/parsed/123/a.xhtml'
        True
        >>> d.storage_policy = "dir"
        >>> d.path('123/a', 'parsed', '.xhtml') == '/tmp/base/parsed/123/a/index.xhtml'
        True
        >>> d.path('123/a', 'downloaded', None, 'r4711', 'appendix.txt') == '/tmp/base/archive/downloaded/123/a/r4711/appendix.txt'
        True
        >>> os.sep = realsep
        
        :param basefile: The basefile for which to calculate the path
        :type  basefile: str
        :param  maindir: The processing stage directory (normally ``downloaded``, ``parsed``, or ``generated``)
        :type   maindin: str
        :param   suffix: The file extension including period (i.e. ``.txt``, not ``txt``)
        :type    suffix: str
        :param  version: Optional, the archived version id
        :type   version: str
        :param attachment: Optional. Any associated file needed by the main file. Requires that ``storage_policy`` is set to ``dir``. ``suffix`` is ignored if this parameter is used.
        :type  attachment: str
        :returns: The full filesystem path
        :rtype:   str

        """
        pathfrag = self.basefile_to_pathfrag(basefile)

        if not storage_policy:
            storage_policy = self.storage_policy
        
        if version:
            v_pathfrag = self.basefile_to_pathfrag(version)
            segments = [self.datadir,
                        'archive', maindir, pathfrag, v_pathfrag]
        else:
            segments = [self.datadir, maindir, pathfrag]

        if storage_policy == "dir":
            if attachment:
                for illegal in ':/':
                    if illegal in attachment:
                        raise errors.AttachmentNameError(
                            "Char '%s' in attachment name '%s' not allowed" % (illegal, attachment))
                segments.append(attachment)
            else:
                segments.append("index" + suffix)
        else:
            if attachment != None:
                raise errors.AttachmentPolicyError(
                    "Can't add attachments (name %s) if "
                    "storage_policy != 'dir'" % attachment)
            segments[-1] += suffix

        unixpath = "/".join(segments)
        if os.sep == "/":
            return unixpath
        else:
            return unixpath.replace("/", os.sep)

    @contextmanager
    def open(self, basefile, maindir, suffix, mode="r", version=None, attachment=None):
        """
        Context manager that opens files for reading or
        writing. The parameters are the same as for :meth:`~ferenda.DocumentStore.path`, and the
        note is applicable here as well -- use
        :meth:`~ferenda.DocumentStore.open_downloaded`, :meth:`~ferenda.DocumentStore.open_parsed` et al if
        possible.

        Example:
        
        >>> store = DocumentStore(datadir="/tmp/base")
        >>> with store.open('123/a', 'parsed', '.xhtml', mode="w") as fp:
        ...     res = fp.write("hello world")
        >>> os.path.exists("/tmp/base/parsed/123/a.xhtml")
        True

        """
        filename = self.path(basefile, maindir, suffix, version, attachment)
        fp = NamedTemporaryFile(mode, delete=False)
        fp.realname = filename
        try:
            yield fp
        finally:
            tempname = fp.name
            fp.close()
            if not os.path.exists(filename) or not filecmp.cmp(tempname, filename):
                util.ensure_dir(filename)
                shutil.move(tempname, filename)
            else:
                os.unlink(tempname)

    def list_basefiles_for(self, action, basedir=None):
        """Get all available basefiles that can be used for the
        specified action.

        :param action: The action for which to get available
                       basefiles (``parse``, ``relate``, ``generate``
                       or ``news``)
        :type action: str
        :param basedir: The base directory in which to search for
                        available files. If not provided, defaults to
                        ``self.datadir``.
        :type basedir: str
        :returns: All available basefiles
        :rtype: generator
        """
        if not basedir:
            basedir = self.datadir
        directory = None
        if action == "parse":
            directory = os.path.sep.join((basedir, "downloaded"))
            if self.storage_policy == "dir":
                # If each document is stored in a separate directory,
                # there is usually other auxillary files (attachments
                # and whatnot) in that directory as well. Make sure we
                # only yield a single file from each directory. By
                # convention, the main file is called index.html,
                # index.pdf or whatever.
                # print("storage_policy dir: %s" % self.storage_policy)
                suffix = os.sep + "index" + self.downloaded_suffix
            else:
                # print("storage_policy file: %s" % self.storage_policy)
                suffix = self.downloaded_suffix
        elif action == "relate":
            directory = os.path.sep.join((basedir, "distilled"))
            suffix = ".rdf"
        elif action == "generate":
            directory = os.path.sep.join((basedir, "parsed"))
            if self.storage_policy == "dir":
                suffix = os.sep + "index.xhtml"
            else:
                suffix = ".xhtml"
        elif action == "news":
            directory = os.path.sep.join((basedir, "entries"))
            suffix = ".json"

        # FIXME: fake action, needed for get_status. replace with
        # something more elegant
        elif action in ("_postgenerate"):
            directory = os.path.sep.join((basedir, "generated"))
            suffix = ".html"

        if not directory:
            raise ValueError("No directory calculated for action %s" % action)

        if not os.path.exists(directory):
            return

        for x in util.list_dirs(directory, suffix, reverse=True):
            # ignore empty files placed by download (which may
            # have done that in order to avoid trying to
            # re-download nonexistent resources)
            if os.path.exists(x) and os.path.getsize(x) > 0:
                # get a pathfrag from full path
                # suffixlen = len(suffix) if self.storage_policy == "file" else len(suffix) + 1
                suffixlen = len(suffix) 
                x = x[len(directory) + 1:-suffixlen]
                yield self.pathfrag_to_basefile(x)

    def list_versions(self, basefile, action=None):
        """Get all archived versions of a given basefile.

        :param basefile: The basefile to list archived versions for
        :type  basefile: str
        :param action: The type of file to look for (either
                       ``downloaded``, ``parsed`` or ``generated``. If
                       ``None``, look for all types.
        :type action: str
        :returns: All available versions for that basefile
        :rtype: generator
        """

        if action:
            assert action in ('downloaded', 'parsed', 'generated'), "Action %s invalid" % action
            actions = (action,)
        else:
            actions = ('downloaded', 'parsed', 'generated')

        basedir = self.datadir
        pathfrag = self.basefile_to_pathfrag(basefile)
        yielded_basefiles = []
        for action in actions:
            directory = os.sep.join((basedir, "archive",
                                     action, pathfrag))
            if not os.path.exists(directory):
                continue
            for x in util.list_dirs(directory, reverse=False):
                if os.path.exists(x):
                    # /datadir/base/archive/downloaded/basefile/version.html
                    # => version.html
                    x = x[len(directory) + 1:]
                    if self.storage_policy == "dir":
                        # version/index.html => version
                        x = os.sep.join(x.split(os.sep)[:-1])
                    else:
                        # version.html => version
                        x = os.path.splitext(x)[0]
                    if os.sep in x:
                        # we didn't find an archived file for
                        # basefile, instead we found an archived file
                        # for another basefile that startswith our
                        # basefile (eg '123' and '123/a', and we found
                        # '123/a/4.html')
                        continue
                    # print("Found file %r %r" % (x, self.pathfrag_to_basefile(x)))
                    basefile = self.pathfrag_to_basefile(x)
                    if basefile not in yielded_basefiles:
                        yielded_basefiles.append(basefile)
                        yield basefile

    def list_attachments(self, basefile, action, version=None):
        """Get all attachments for a basefile in a specified state

        :param action: The state (type of file) to look for (either
                       ``downloaded``, ``parsed`` or ``generated``. If
                       ``None``, look for all types.
        :type action: str
        :param basefile: The basefile to list attachments for
        :type  basefile: str
        :param version: The version of the basefile to list attachments for. If None, list attachments for the current version.
        :type  version: str
        :returns: All available attachments for the basefile 
        :rtype: generator
        """
        if self.storage_policy != "dir":
            raise errors.AttachmentPolicyError("Can't list attachments if storage_policy != 'dir'")
        
        basedir = self.datadir
        # pathfrag = self.pathfrag_to_basefile(basefile) # that can't be right?
        pathfrag = self.basefile_to_pathfrag(basefile)
        if version:
            v_pathfrag = self.basefile_to_pathfrag(version)
            directory = os.sep.join((basedir, "archive", action, pathfrag, v_pathfrag))
        else:
            directory = os.sep.join((basedir, action, pathfrag))
        # FIXME: Similar map exists in list_basefiles_for and in other
        # places throughout the code. Should subclasses be able to
        # control suffixes beyond the simple self.downloaded_suffix
        # mechanism?
        suffixmap = {'downloaded': self.downloaded_suffix,
                     'parsed': '.xhtml',
                     'generated': '.html'}
        mainfile = "index" + suffixmap[action]
        for x in util.list_dirs(directory, reverse=False):
            # /datadir/base/downloaded/basefile/attachment.txt => attachment.txt
            x = x[len(directory) + 1:]
            if x != mainfile:
                yield x

    def basefile_to_pathfrag(self, basefile):
        """Given a basefile, returns a string that can safely be used
        as a fragment of the path for any representation of that
        file. The default implementation recognizes a number of
        characters that are unsafe to use in file names and replaces
        them with HTTP percent-style encoding.

        Example:

        >>> d = DocumentStore("/tmp")
        >>> realsep = os.sep
        >>> os.sep = "/"
        >>> d.basefile_to_pathfrag('1998:204') == '1998/%3A204'
        True
        >>> os.sep = realsep

        If you wish to override how document files are stored in
        directories, you can override this method, but you should make
        sure to also override
        :py:meth:`~ferenda.DocumentStore.pathfrag_to_basefile` to
        work as the inverse of this method.

        :param basefile: The basefile to encode
        :type basefile: str
        :returns: The encoded path fragment
        :rtype: str
        """
        safe = '/;@&=+,'

        if six.PY2:
            # urllib.quote in python 2 cannot handle unicode values
            # for the s parameter (2.6 cannot even handle unicode
            # values for the safe parameter). FIXME: We should create
            # a shim as ferenda.compat.quote and use that
            basefile = basefile.encode('utf-8')
            safe = safe.encode('ascii') # pragma: no cover
            
        return quote(basefile, safe=safe).replace('%', os.sep + '%')

    def pathfrag_to_basefile(self, pathfrag):
        """Does the inverse of
        :py:meth:`~ferenda.DocumentStore.basefile_to_pathfrag`,
        that is, converts a fragment of a file path into the
        corresponding basefile.

        :param pathfrag: The path fragment to decode
        :type pathfrag: str
        :returns: The resulting basefile
        :rtype: str
        """
        if os.sep == "\\":
            pathfrag = pathfrag.replace("\\", "/")
        return unquote(pathfrag.replace('/%', '%'))

    def archive(self, basefile, version):
        """Moves the current version of a document to an archive. All
        files related to the document are moved (downloaded, parsed,
        generated files and any existing attachment files).

        :param basefile: The basefile of the document to archive
        :type basefile: str
        :param version: The version id to archive under
        :type version: str
        """

        for meth in (self.downloaded_path, self.documententry_path,
                     self.parsed_path, self.serialized_path,
                     self.distilled_path,
                     self.annotation_path, self.generated_path):
            # FIXME: what about intermediate? Ignore them as they
            # should be able to be regenerated at any time?
            src = meth(basefile)
            dest = meth(basefile, version)
            if self.storage_policy == "dir" and meth in (self.downloaded_path,
                                                         self.parsed_path,
                                                         self.generated_path):
                src = os.path.dirname(src)
                dest = os.path.dirname(dest)
            if not os.path.exists(src):
                continue
            if os.path.exists(dest):
                raise errors.ArchivingError(
                    "Archive destination %s for basefile %s version %s already exists!" % (dest, basefile, version))
            # self.log.debug("Archiving %s to %s" % (src,dest))
            # print("Archiving %s to %s" % (src,dest))
            util.ensure_dir(dest)
            shutil.move(src, dest)

    def downloaded_path(self, basefile, version=None, attachment=None):
        """Get the full path for the downloaded file for the given
        basefile (and optionally archived version and/or attachment
        filename).

        :param basefile: The basefile for which to calculate the path
        :type  basefile: str
        :param  version: Optional. The archived version id
        :type   version: str
        :param attachment: Optional. Any associated file needed by the main file. 
        :type  attachment: str
        :returns: The full filesystem path
        :rtype:   str
        """
        return self.path(basefile, 'downloaded',
                         self.downloaded_suffix, version, attachment)

    def open_downloaded(self, basefile, mode="r", version=None, attachment=None):
        """Opens files for reading and writing,
        c.f. :meth:`~ferenda.DocumentStore.open`. The parameters are
        the same as for
        :meth:`~ferenda.DocumentStore.downloaded_path`.

        """

        filename = self.downloaded_path(basefile, version, attachment)
        return self._open(filename, mode)

    def documententry_path(self, basefile, version=None):
        """Get the full path for the documententry JSON file for the given
        basefile (and optionally archived version).

        :param basefile: The basefile for which to calculate the path
        :type  basefile: str
        :param  version: Optional. The archived version id
        :type   version: str
        :returns: The full filesystem path
        :rtype:   str
        """
        return self.path(basefile, 'entries', '.json', version, storage_policy="file")

    def intermediate_path(self, basefile, version=None, attachment=None):
        """Get the full path for the main intermediate file for the given
        basefile (and optionally archived version).

        :param basefile: The basefile for which to calculate the path
        :type  basefile: str
        :param  version: Optional. The archived version id
        :type   version: str
        :param attachment: Optional. Any associated file created or retained
                           in the intermediate step
        :returns: The full filesystem path
        :rtype:   str
        """
        return self.path(basefile, 'intermediate', '.xml', version, attachment)

    def open_intermediate(self, basefile, mode="r", version=None, attachment=None):
        """Opens files for reading and writing,
        c.f. :meth:`~ferenda.DocumentStore.open`. The parameters are
        the same as for
        :meth:`~ferenda.DocumentStore.intermediate_path`.
        """
        filename = self.intermediate_path(basefile, version, attachment)
        return self._open(filename, mode)

    def parsed_path(self, basefile, version=None, attachment=None):
        """Get the full path for the parsed XHTML file for the given
        basefile.

        :param basefile: The basefile for which to calculate the path
        :type  basefile: str
        :param  version: Optional. The archived version id
        :type   version: str
        :param attachment: Optional. Any associated file needed by the
                           main file (created by
                           :py:meth:`~ferenda.DocumentRepository.parse`)
        :type  attachment: str
        :returns: The full filesystem path
        :rtype:   str
        """
        return self.path(basefile, 'parsed', '.xhtml',
                         version, attachment)

    def open_parsed(self, basefile, mode="r", version=None, attachment=None):
        """Opens files for reading and writing,
        c.f. :meth:`~ferenda.DocumentStore.open`. The parameters are
        the same as for
        :meth:`~ferenda.DocumentStore.parsed_path`.

        """
        filename = self.parsed_path(basefile, version, attachment)
        return self._open(filename, mode)

    def serialized_path(self, basefile, version=None, attachment=None):
        """Get the full path for the serialized JSON file for the given
        basefile.

        :param basefile: The basefile for which to calculate the path
        :type  basefile: str
        :param  version: Optional. The archived version id
        :type   version: str
        :returns: The full filesystem path
        :rtype:   str
        """
        return self.path(basefile, 'serialized', '.json',
                         version, storage_policy="file")

    def open_serialized(self, basefile, mode="r", version=None):
        """Opens files for reading and writing,
        c.f. :meth:`~ferenda.DocumentStore.open`. The parameters are
        the same as for
        :meth:`~ferenda.DocumentStore.serialized_path`.

        """
        filename = self.serialized_path(basefile, version)
        return self._open(filename, mode)

    def distilled_path(self, basefile, version=None):
        """Get the full path for the distilled RDF/XML file for the given
        basefile.

        :param basefile: The basefile for which to calculate the path
        :type  basefile: str
        :param  version: Optional. The archived version id
        :type   version: str
        :returns: The full filesystem path
        :rtype:   str
        """
        return self.path(basefile, 'distilled', '.rdf',
                         version, storage_policy="file")

    def open_distilled(self, basefile, mode="r", version=None):
        """Opens files for reading and writing,
        c.f. :meth:`~ferenda.DocumentStore.open`. The parameters are
        the same as for
        :meth:`~ferenda.DocumentStore.distilled_path`.

        """
        filename = self.distilled_path(basefile, version)
        return self._open(filename, mode)

    def generated_path(self, basefile, version=None, attachment=None):
        """Get the full path for the generated file for the given
        basefile (and optionally archived version and/or attachment
        filename).

        :param basefile: The basefile for which to calculate the path
        :type  basefile: str
        :param  version: Optional. The archived version id
        :type   version: str
        :param attachment: Optional. Any associated file needed by the main file. 
        :type  attachment: str
        :returns: The full filesystem path
        :rtype:   str
        """
        return self.path(basefile, 'generated', '.html',
                         version, attachment)

# Removed this method until I find a reason to use it
#
#    def open_generated(self, basefile, mode="r", version=None, attachment=None):
#        """Opens files for reading and writing,
#        c.f. :meth:`~ferenda.DocumentStore.open`. The parameters are
#        the same as for
#        :meth:`~ferenda.DocumentStore.generated_path`.
#
#        """
#        filename = self.generated_path(basefile, version, attachment)
#        return self._open(filename, mode)

    def annotation_path(self, basefile, version=None):
        """Get the full path for the annotation file for the given
        basefile (and optionally archived version).

        :param basefile: The basefile for which to calculate the path
        :type  basefile: str
        :param  version: Optional. The archived version id
        :type   version: str
        :returns: The full filesystem path
        :rtype:   str
        """
        return self.path(basefile, 'annotations', '.grit.xml',
                         version, storage_policy="file")

    def open_annotation(self, basefile, mode="r", version=None):
        """Opens files for reading and writing,
        c.f. :meth:`~ferenda.DocumentStore.open`. The parameters are
        the same as for
        :meth:`~ferenda.DocumentStore.annotation_path`."""
        filename = self.annotation_path(basefile, version)
        return self._open(filename, mode)

    def dependencies_path(self, basefile):
        """Get the full path for the dependency file for the given
        basefile

        :param basefile: The basefile for which to calculate the path
        :type  basefile: str
        :returns: The full filesystem path
        :rtype:   str
        """
        return self.path(basefile, 'deps', '.txt', storage_policy="file")

    def open_dependencies(self, basefile, mode="r"):
        """Opens files for reading and writing,
        c.f. :meth:`~ferenda.DocumentStore.open`. The parameters are
        the same as for
        :meth:`~ferenda.DocumentStore.dependencies_path`."""
        filename = self.dependencies_path(basefile)
        return self._open(filename, mode)

    def atom_path(self, basefile):
        """Get the full path for the atom file for the given
        basefile

        .. note::

           This is used by :meth:`ferenda.DocumentRepository.news` and
           does not really operate on "real" basefiles. It might be
           removed. You probably shouldn't use it unless you override
           :meth:`~ferenda.DocumentRepository.news`

        :param basefile: The basefile for which to calculate the path
        :type  basefile: str
        :returns: The full filesystem path
        :rtype:   str

        """
        return self.path(basefile, 'feed', '.atom', storage_policy="file")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re
from collections import deque, defaultdict

from rdflib import URIRef

from ferenda import DocumentRepository
from ferenda.elements import CompoundElement, OrdinalElement

# More TODO: create test/files/repo/eut/source/all.json like
# {'http://eur-lex.europa.eu/LexUriServ/LexUriServ.do?uri=OJ:C:2008:115:0001:01:EN:HTML': 'treaties.html'}
#
# Create downloaded/tfeu.html and teu.html (same content or possibly a bit shortened)
#
# Create distilled/tfeu.ttl and distilled/teu.ttl, should include structural elements
#
# Create parsed/tfeu.xhtml and parsed/teu.xhtml.
#
# Then done!

# The general outline of a treaty is:
# <Body> C
#   <Paragraph> C (unicode/Link) - starting and ending titles
#   <Preamble> C
#     <Paragraph> - the typographic term, aka "Stycke"
#   <Part> CO - not present for TEU
#     <Title> CO
#       <Chapter> CO
#         <Section> CO
#           <Article> CO
#             <Subarticle> CO
#                <Paragraph> C
#                  <unicode>
#                  <Link>
#                <UnordedList leader="dash"> C
#                  <ListItem> C
#                <OrderedList type="letter"> CO

# or should we have a class method ontology_uri, complimentary to canonical_uri/dataset_uri ?
vocab_uri = "http://lagen.nu/eurlex#"


class PreambleRecital(CompoundElement, OrdinalElement):
    pass

# the most toplevel structural element. Only used for TFEU, not TEU


class Part(CompoundElement, OrdinalElement):
    pass

# nb: this is completely different from ferenda.elements.Title -- this title is a toplevel
# structural element that encompasses chapters, sections, articles etc


class Title(CompoundElement, OrdinalElement):
    pass


class Chapter(CompoundElement, OrdinalElement):
    pass


class Section(CompoundElement, OrdinalElement):
    pass


class Article(CompoundElement, OrdinalElement):
    fragment_label = "A"
    # FIXME: extend CompoundElement.as_xhtml to check for rdf_type and use it as an @about
    # attribute (using a make_graph() graph to qname it)
    rdf_type = URIRef(vocab_uri + "Article")


class Subarticle(CompoundElement, OrdinalElement):
    fragment_label = "P"
    rdf_type = URIRef(vocab_uri + "SubArticle")


class ListItem(CompoundElement):
    fragment_label = "L"
    rdf_type = URIRef(vocab_uri + "ListItem")


class EurlexTreaties(DocumentRepository):

    """Handles the foundation treaties of the European union."""
    # overrides of superclass variables
    alias = "eut"  # European Union Treaties
    start_url = "http://eur-lex.europa.eu/LexUriServ/LexUriServ.do?uri=OJ:C:2008:115:0001:01:EN:HTML"
    document_url_template = "http://eur-lex.europa.eu/LexUriServ/LexUriServ.do?uri=OJ:C:2008:115:0001:01:EN:HTML#%(basefile)s"
    rdf_type = URIRef(vocab_uri + "Treaty")

#
# Downloading

    def download(self, basefile=None):
        # NB: The very same document contains both TEU and TFEU. We download it twice
        # (wasting some storage space) and let parse() pick out the relevant parts.
        self.download_single("teu")
        self.download_single("tfeu")

#
# Parsing -- FIXME: this should be easily ported to FSMParser

    re_part = re.compile("PART (ONE|TWO|THREE|FOUR|FIVE|SIX|SEVEN)$").match
    re_title = re.compile("TITLE ([IVX]+)$").match
    re_chapter = re.compile("CHAPTER (\d+)$").match
    re_section = re.compile("SECTION (\d+)$").match
    re_article = re.compile("Article (\d+)$").match
    re_subarticle = re.compile("^(\d+)\. ").search
    re_unorderedliststart = re.compile("^- ").search
    re_orderedliststart = re.compile("^\(\w\) ").search
    re_romanliststart = re.compile("^\([ivx]+\) ").search
    ordinal_list = ('ONE', 'TWO', 'THREE', 'FOUR', 'FIVE', 'SIX', 'SEVEN',
                    'EIGHT', 'NINE', 'TEN', 'ELEVEN', 'TWELVE')
    ordinal_dict = dict(
        list(zip(ordinal_list, list(range(1, len(ordinal_list) + 1)))))

    def parse_metadata_from_soup(self, soup, doc):
        if not doc.uri:
            doc.uri = self.canonical_uri(doc.basefile)
        desc = Describer(doc.meta, doc.uri)
        desc.rdftype(self.rdf_type)
        if basefile == "teu":
            desc.value(self.ns['dct'].title, "Treaty on European Union", lang="en")
        elif basefile == "tfeu":
            desc.value(
                self.ns['dct'].title, "Treaty on the Functioning of the European Union", lang="en")

    def parse_document_from_soup(soup, doc):
        if basefile == "teu":
            startnode = soup.findAll(text="-" * 50)[1].parent
        elif basefile == "tfeu":
            startnode = soup.findAll(text="-" * 50)[2].parent

        lines = deque()
        for p in startnode.findNextSiblings("p"):
            if p.string == "-" * 50:
                self.log.info("found the end")
                break
            else:
                if p.string:
                    lines.append(str(p.string))

        doc.body = self.make_body(lines)
        self.process_body(doc.body, '', doc.uri)

    def make_body(self, lines):
        b = Body()
        while lines:
            line = lines.popleft()
            if line == "PREAMBLE":
                b.append(self.make_preamble(lines))
            elif self.re_title(line):
                lines.appendleft(line)
                b.append(self.make_title(lines))
            elif self.re_part(line):
                lines.appendleft(line)
                b.append(self.make_part(lines))
            else:
                b.append(Paragraph([line]))
            # print type(b[-1])
        return b

    def make_preamble(self, lines):
        p = PreambleRecital(title="PREAMBLE")
        while lines:
            line = lines.popleft()
            if (self.re_part(line) or self.re_title(line)):
                lines.appendleft(line)
                return p
            else:
                p.append(Paragraph([line]))

        self.log.warn("make_preamble ran out of lines!")
        return p

    def make_part(self, lines):
        partnumber = lines.popleft()
        ordinal = self.ordinal_dict[self.re_part(partnumber).group(1)]
        parttitle = lines.popleft()
        p = Part(ordinal=ordinal, ordinaltitle=partnumber, title=parttitle)
        while lines:
            line = lines.popleft()
            if (self.re_part(line)):
                lines.appendleft(line)
                return p
            elif (self.re_title(line)):
                lines.appendleft(line)
                p.append(self.make_title(lines))
            elif (self.re_article(line)):
                # print "make_part: %s matches article" % line
                lines.appendleft(line)
                p.append(self.make_article(lines))
            else:
                p.append(Paragraph([line]))
                self.log.warn(
                    "make_part appended naked Paragraph '%s...'" % line[:25])
        return p

    def make_title(self, lines):
        titlenumber = lines.popleft()
        ordinal = self._from_roman(self.re_title(titlenumber).group(1))
        titletitle = lines.popleft()
        t = Title(ordinal=ordinal, ordinaltitle=titlenumber, title=titletitle)
        while lines:
            line = lines.popleft()
            if (self.re_part(line) or self.re_title(line)):
                lines.appendleft(line)
                return t
            elif (self.re_chapter(line)):
                lines.appendleft(line)
                t.append(self.make_chapter(lines))
            elif (self.re_article(line)):
                # print "make_title: %s matches article" % line
                lines.appendleft(line)
                t.append(self.make_article(lines))
            else:
                t.append(Paragraph([line]))
                self.log.warn(
                    "make_title appended naked Paragraph '%s...'" % line[:25])
        return t

    def make_chapter(self, lines):
        chapternumber = lines.popleft()
        ordinal = int(self.re_chapter(chapternumber).group(1))
        chaptertitle = lines.popleft()
        c = Chapter(
            ordinal=ordinal, ordinaltitle=chapternumber, title=chaptertitle)
        while lines:
            line = lines.popleft()
            if (self.re_part(line) or
                self.re_title(line) or
                    self.re_chapter(line)):
                lines.appendleft(line)
                return c
            elif (self.re_section(line)):
                lines.appendleft(line)
                c.append(self.make_section(lines))
            elif (self.re_article(line)):
                # print "make_chapter: %s matches article" % line
                lines.appendleft(line)
                c.append(self.make_article(lines))
            else:
                c.append(Paragraph([line]))
                self.log.warn("make_chapter appended naked Paragraph '%s...'" %
                              line[:25])
        return c

    def make_section(self, lines):
        sectionnumber = lines.popleft()
        ordinal = int(self.re_section(sectionnumber).group(1))
        sectiontitle = lines.popleft()
        s = Section(
            ordinal=ordinal, ordinaltitle=sectionnumber, title=sectiontitle)
        while lines:
            line = lines.popleft()
            if (self.re_part(line) or
                self.re_title(line) or
                self.re_chapter(line) or
                    self.re_section(line)):
                lines.appendleft(line)
                return s
            elif (self.re_article(line)):
                # print "make_section: %s matches article" % line
                lines.appendleft(line)
                s.append(self.make_article(lines))
            else:
                s.append(Paragraph([line]))
                self.log.warn("make_section appended naked Paragraph '%s...'" %
                              line[:25])
        return s

    def make_article(self, lines):
        articlenumber = lines.popleft()
        ordinal = int(self.re_article(articlenumber).group(1))
        self.log.info("Making article: %s" % ordinal)
        exarticlenumber = lines.popleft()
        if not exarticlenumber.startswith("(ex Article"):
            lines.appendleft(exarticlenumber)
            a = Article(ordinal=ordinal, ordinaltitle=articlenumber)
        else:
            a = Article(ordinal=ordinal, ordinaltitle=articlenumber,
                        exarticlenumber=exarticlenumber)

        while lines:
            line = lines.popleft()
            if (self.re_part(line) or
                self.re_title(line) or
                self.re_chapter(line) or
                self.re_section(line) or
                    self.re_article(line)):
                lines.appendleft(line)
                return a
            elif (self.re_subarticle(line)):
                lines.appendleft(line)
                a.append(self.make_subarticle(lines))
            elif (self.re_unorderedliststart(line)):
                lines.appendleft(line)
                a.append(self.make_unordered_list(lines, "dash"))
            elif (self.re_orderedliststart(line)):
                lines.appendleft(line)
                a.append(self.make_ordered_list(lines, "lower-alpha"))
            else:
                # print "Appending %s" % line[:40]
                a.append(Paragraph([line]))

        return a

    def make_subarticle(self, lines):
        line = lines.popleft()
        subarticlenum = int(self.re_subarticle(line).group(1))
        # self.log.info("Making subarticle %d: %s" % (subarticlenum, line[:30]))
        s = Subarticle(ordinal=subarticlenum)
        lines.appendleft(line)
        while lines:
            line = lines.popleft()
            if (self.re_part(line) or
                self.re_title(line) or
                self.re_chapter(line) or
                self.re_section(line) or
                    self.re_article(line)):
                lines.appendleft(line)
                return s
            elif (self.re_subarticle(line) and
                  int(self.re_subarticle(line).group(1)) != subarticlenum):
                lines.appendleft(line)
                return s
            elif (self.re_unorderedliststart(line)):
                lines.appendleft(line)
                s.append(self.make_unordered_list(lines, "dash"))
            elif (self.re_orderedliststart(line)):
                lines.appendleft(line)
                s.append(self.make_ordered_list(lines, "lower-alpha"))
            else:
                # this is OK
                s.append(Paragraph([line]))
        return s

    def make_unordered_list(self, lines, style):
        ul = UnorderedList(style=style)
        while lines:
            line = lines.popleft()
            if not self.re_unorderedliststart(line):
                lines.appendleft(line)
                return ul
            else:
                ul.append(ListItem([line]))
        return ul

    def make_ordered_list(self, lines, style):
        ol = OrderedList(style=style)
        while lines:
            line = lines.popleft()
            # try romanliststart before orderedliststart -- (i) matches
            # both, but is likely the former
            if self.re_romanliststart(line):
                # print "make_ordered_list: re_romanliststart: %s" % line[:40]
                if style == "lower-roman":
                    ol.append(ListItem([line]))
                else:
                    lines.appendleft(line)
                    ol.append(self.make_ordered_list(lines, "lower-roman"))
            elif self.re_orderedliststart(line):
                # print "make_ordered_list: re_orderedliststart: %s" % line[:40]
                if style == "lower-alpha":
                    ol.append(ListItem([line]))
                else:  # we were in a roman-style sublist, so we should pop up
                    lines.appendleft(line)
                    return ol
            else:
                # print "make_ordered_list: done: %s" % line[:40]
                lines.appendleft(line)
                return ol
        return ol

    # Post-process the document tree in a recursive fashion in order to:
    #
    # Find addressable units (resources that should have unique URI:s,
    # e.g. articles and subarticles) and construct IDs for them, like
    # "A7", "A25(b)(ii)" (or A25S1P2N2 or...?)
    #
    # How should we handle Articles themselves -- they have individual
    # CELEX numbers and therefore URIs (but subarticles don't)?
    def process_body(self, element, prefix, baseuri):
        if isinstance(element, str):
            return
        # print "Starting with "  + str(type(element))
        counters = defaultdict(int)
        for p in element:
            counters[type(p)] += 1
            # print "handling " + str(type(p))
            if hasattr(p, 'fragment_label'):  # this is an addressable resource
                elementtype = p.fragment_label
                if hasattr(p, 'ordinal'):
                    elementordinal = p.ordinal
                else:
                    elementordinal = counters[type(p)]

                fragment = "%s%s%s" % (prefix, elementtype, elementordinal)
                if elementtype == "A":
                    uri = "%s%03d" % (baseuri, elementordinal)
                else:
                    uri = "%s%s%s" % (baseuri, elementtype, elementordinal)

                p.id = fragment
                p.attrs = {'id': p.id,
                           'about': uri,
                           'typeof': p.rdftype}
                if elementtype == "A":
                    uri += "#"
            else:
                fragment = prefix
                uri = baseuri

            self.process_body(p, fragment, uri)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals


class TocPageset(object):

    """Represents a particular set of TOC pages, structured around some
particular attribute(s) of documents, like title or publication
date. :py:meth:`~ferenda.DocumentRepository.toc_pagesets` returns a
list of these objects, override that method to provide custom
TocPageset objects.

    :param label: A description of this set of TOC pages, like "By publication year"
    :type  label: str
    :param pages: The set of :py:class:`~ferenda.TocPage` objects that makes up this page set.
    :type  pages: list

    """

    def __init__(self, label, pages):
        self.label = label
        self.pages = pages

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        dictrepr = "".join((" %s=%s" % (k, v) for k, v in sorted(self.__dict__.items())))
        return ("<%s%s>" % (self.__class__.__name__, dictrepr))



# -*- coding: utf-8 -*-
"""These are the exceptions thrown by Ferenda. Any of the python built-in exceptions may be thrown as well, but exceptions in used third-party libraries should be wrapped in one of these."""
from __future__ import unicode_literals


class ParseError(Exception):

    """Raised when :py:meth:`~ferenda.DocumentRepository.parse` fails in
    any way.
    """


class FSMStateError(ParseError):

    """Raised whenever the current state and the current symbol in a
       :py:class:`~ferenda.FSMParser` configuration does not have a
       defined transition.
    """


class DocumentRemovedError(Exception):

    """Raised whenever a particular document has been found to be removed
    -- this can happen either during
    :py:meth:`~ferenda.DocumentRepository.download` or
    :py:meth:`~ferenda.DocumentRepository.parse` (which may be the
    case if there exists a physical document, but whose contents are
    essentially a placeholder saying that the document has been
    removed).

    You can set the attribute ``dummyfile`` on this exception when
    raising it, preferably to the parsed_path that would be created,
    if not this exception had occurred.. If present,
    ``ferenda-build.py`` (or rather :meth:`ferenda.manager.run`) will
    use this to create a dummy file at the indicated path. This
    prevents endless re-parsing of expired documents.

    """


class PatchError(ParseError):
    """Raised if a patch cannot be applied by
:py:meth:`~ferenda.DocumentRepository.patch_if_needed`."""

class NoDownloadedFileError(ParseError):
    """Raised on an attempt to parse a basefile for which there doesn't
exist a downloaded file.""" 
    
    
class AttachmentNameError(ValueError):

    """Raised whenever an invalid attachment name is used with any method
    of :py:class:`~ferenda.DocumentStore`."""


class AttachmentPolicyError(ValueError):

    """Raised on any attempt to store an attachment using
    :py:class:`~ferenda.DocumentStore` when ``storage_policy`` is not
    set to ``dir``.
    """


class ArchivingError(Exception):

    """Raised whenever an attempt to archive a document version using :py:meth:`~ferenda.DocumentStore.archive` fails (for example, because the archive version 
already exists)."""


class ValidationError(Exception):

    """Raised whenever a created document doesn't validate using the
    appropriate schema."""


class TransformError(Exception):

    """Raised whenever a XSLT transformation fails for any reason."""


class ExternalCommandError(Exception):

    """Raised whenever any invocation of an external commmand fails for
    any reason (including if the command line program doesn't exist)."""


class ConfigurationError(Exception):

    """Raised when a configuration file cannot be found in it's expected
location or when it cannot be used due to corruption, file permissions
or other reasons"""


class TriplestoreError(Exception):

    """Raised whenever communications with the triple store fails, for whatever reason."""


class SparqlError(TriplestoreError):

    """Raised whenever a SPARQL query fails. The Exception should contain whatever error message that the Triple store returned, so the exact formatting may be dependent on which store is used."""


class IndexingError(Exception):

    """Raised whenever an attempt to put text into the fulltext index fails."""


class SearchingError(Exception):

    """Raised whenever an attempt to do a full-text search fails."""



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sys, os
from ferenda.compat import unittest
if os.getcwd() not in sys.path: sys.path.insert(0,os.getcwd())

import rdflib

#SUT
from ferenda import Document

class TestDocument(unittest.TestCase):
    def test_init(self):
        d = Document()
        self.assertIsInstance(d.meta, rdflib.Graph)
        self.assertEqual(d.body, [])
        self.assertIsNone(d.uri)
        self.assertIsNone(d.lang)
        self.assertIsNone(d.basefile)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re
import os
from datetime import datetime
import time
import itertools

from six import text_type as str
from six import binary_type as bytes

# 3rd party
from bs4 import BeautifulSoup
import requests
import requests.exceptions
from rdflib import URIRef, Literal

# My own stuff
from ferenda import PDFDocumentRepository, DocumentStore, PDFReader, WordReader, Describer
from ferenda import util
from ferenda.decorators import downloadmax, recordlastdownload, managedparsing
from ferenda.elements import Body, UnicodeElement, CompoundElement, serialize
from . import SwedishLegalSource, SwedishCitationParser, RPUBL
from ferenda.sources.legal.se.legalref import LegalRef


class ARNStore(DocumentStore):
    """Customized DocumentStore."""
    def basefile_to_pathfrag(self, basefile):
        return basefile.replace("-", "/")

    def pathfrag_to_basefile(self, pathfrag):
        return pathfrag.replace("/","-", 1)

    def downloaded_path(self, basefile, version=None, attachment=None, suffix=None):
        if not suffix:
            if os.path.exists(self.path(basefile, "downloaded", ".wpd")):
                suffix = ".wpd"
            elif os.path.exists(self.path(basefile, "downloaded", ".doc")):
                suffix = ".doc"
            elif os.path.exists(self.path(basefile, "downloaded", ".docx")):
                suffix = ".docx"
            elif os.path.exists(self.path(basefile, "downloaded", ".rtf")):
                suffix = ".rtf"
            elif os.path.exists(self.path(basefile, "downloaded", ".pdf")):
                suffix = ".pdf"
            else:
                suffix = self.downloaded_suffix
        return self.path(basefile, "downloaded", suffix, version, attachment)

    def list_basefiles_for(self, action, basedir=None):
        if not basedir:
            basedir = self.datadir
        if action == "parse":
            d = os.path.sep.join((basedir, "downloaded"))
            if not os.path.exists(d):
                return
            for x in sorted(itertools.chain(util.list_dirs(d, ".wpd"),
                                            util.list_dirs(d, ".rtf"),
                                            util.list_dirs(d, ".doc"),
                                            util.list_dirs(d, ".docx"),
                                            util.list_dirs(d, ".pdf"))):
                suffix = "/index"+ os.path.splitext(x)[1]
                pathfrag = x[len(d) + 1:-len(suffix)]
                yield self.pathfrag_to_basefile(pathfrag)
        else:
            for x in super(ARNStore, self).list_basefiles_for(action, basedir):
                yield x

class ARN(SwedishLegalSource, PDFDocumentRepository):

    """Hanterar referat från Allmänna Reklamationsnämnden, www.arn.se.

    Modulen hanterar hämtande av referat från ARNs webbplats, omvandlande
    av dessa till XHTML1.1+RDFa, samt transformering till browserfärdig
    HTML5.
    """

    alias = "arn"
    xslt_template = "res/xsl/arn.xsl"
    start_url = "http://adokweb.arn.se/digiforms/sessionInitializer?processName=SearchRefCasesProcess"
    documentstore_class = ARNStore
    rdf_type = RPUBL.VagledandeMyndighetsavgorande
    
    @recordlastdownload
    def download(self, basefile=None):
        self.session = requests.Session()
        resp = self.session.get(self.start_url)
        soup = BeautifulSoup(resp.text)
        action = soup.find("form")["action"]

        if self.config.lastdownload and not self.config.refresh:
            d = self.config.lastdownload
            datefrom = '%d-%02d-%02d' % (d.year, d.month, d.day)
            dateto = '%d-01-01' % (d.year+1)
        else:
            # only fetch one year at a time
            datefrom = '1992-01-01'
            dateto = '1993-01-01'
        
        params = {
            '/root/searchTemplate/decision': 'obegransad',
            '/root/searchTemplate/decisionDateFrom': datefrom,
            '/root/searchTemplate/decisionDateTo': dateto,
            '/root/searchTemplate/department': 'alla',
            '/root/searchTemplate/journalId': '',
            '/root/searchTemplate/searchExpression': '',
            '_cParam0': 'method=search',
            '_cmdName': 'cmd_process_next',
            '_validate': 'page'
        }

        for basefile, url, fragment in self.download_get_basefiles((action, params)):
            if (self.config.refresh or
                    (not os.path.exists(self.store.downloaded_path(basefile)))):
                self.download_single(basefile, url, fragment)

    @downloadmax
    def download_get_basefiles(self, args):
        action, params = args
        done = False
        self.log.debug("Retrieving all results from %s to %s" %
                       (params['/root/searchTemplate/decisionDateFrom'],
                        params['/root/searchTemplate/decisionDateTo']))
        paramcopy = dict(params)
        while not done:
            # First we need to use the files argument to send the POST
            # request as multipart/form-data
            req = requests.Request(
                "POST", action, cookies=self.session.cookies, files=paramcopy).prepare()
            # Then we need to remove filename
            # from req.body in an unsupported manner in order not to
            # upset the sensitive server
            body = req.body
            if isinstance(body, bytes):
                body = body.decode() # should be pure ascii
            req.body = re.sub(
                '; filename="[\w\-\/]+"', '', body).encode()
            req.headers['Content-Length'] = str(len(req.body))
            # And finally we have to allow RFC-violating redirects for POST

            resp = False
            remaining_attempts = 5
            while (not resp) and (remaining_attempts > 0):
                try:
                    resp = self.session.send(req, allow_redirects=True)
                except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as e:
                    self.log.warning(
                        "Failed to POST %s: error %s (%s remaining attempts)" % (action, e, remaining_attempts))
                    remaining_attempts -= 1
                    time.sleep(1)
                        
            soup = BeautifulSoup(resp.text)
            for link in soup.find_all("input", "standardlink", onclick=re.compile("javascript:window.open")):
                url = link['onclick'][24:-2]  # remove 'javascript:window.open' call around the url
                # this probably wont break...
                fragment = link.find_parent("table").find_parent("table")
                basefile = fragment.find_all("div", "strongstandardtext")[1].text
                yield basefile, url, fragment
            if soup.find("input", value="Nästa sida"):
                self.log.debug("Now retrieving next page in current search")
                paramcopy = {'_cParam0': "method=nextPage",
                             '_validate': "none",
                             '_cmdName': "cmd_process_next"}
            else:
                fromYear = int(params['/root/searchTemplate/decisionDateFrom'][:4])
                if fromYear >= datetime.now().year:
                    done = True
                else:
                    # advance one year
                    params['/root/searchTemplate/decisionDateFrom'] = "%s-01-01" % str(fromYear+1)
                    params['/root/searchTemplate/decisionDateTo'] = "%s-01-01" % str(fromYear+2)
                    self.log.debug("Now retrieving all results from %s to %s" %
                                   (params['/root/searchTemplate/decisionDateFrom'],
                                    params['/root/searchTemplate/decisionDateTo']))
                    paramcopy = dict(params)
                    # restart the search, so that poor digiforms
                    # doesn't get confused

                    resp = False
                    remaining_attempts = 5
                    while (not resp) and (remaining_attempts > 0):
                        try:
                            resp = self.session.get(self.start_url)
                        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as e:
                            self.log.warning(
                                "Failed to POST %s: error %s (%s remaining attempts)" % (action, e, remaining_attempts))
                            remaining_attempts -= 1
                            time.sleep(1)

                    soup = BeautifulSoup(resp.text)
                    action = soup.find("form")["action"]

    
    def guess_type(self, filename):
        with open(filename, "rb") as fp:
            sig = fp.read(4)
            if sig == b'\xffWPC':
                doctype = ".wpd"
            elif sig == b'\xd0\xcf\x11\xe0':
                doctype = ".doc"
            elif sig == b'PK\x03\x04':
                doctype = ".docx"
            elif sig == b'{\\rt':
                doctype = ".rtf"
            elif sig == b'%PDF':
                doctype = ".pdf"
            else:
                self.log.warning(
                    "%s has unknown signature %r -- don't know what kind of file it is" % (filename, sig))
                doctype = ".pdf"  # don't do anything
        return doctype
        
                    
    def download_single(self, basefile, url, fragment):
        ret = super(ARN, self).download_single(basefile, url)
        # after downloading: see if our PDF in reality was something else
        # FIXME: we should do this prior to .download_if_needed...
        if ret:
            d = self.store.downloaded_path(basefile)
            if os.path.exists(d) and not d.endswith(".pdf"):
                doctype = self.guess_type(d)
                if doctype != '.pdf':
                    util.robust_rename(d, d.replace(".pdf", doctype))

            # the HTML fragment from the search result page contains
            # metadata not available in the main document, so save it
            # as fragment.html
            with self.store.open_downloaded(basefile, mode="w", attachment="fragment.html") as fp:
                fp.write(str(fragment))
        return ret
        
    @managedparsing
    def parse(self, doc):
        def nextcell(key):
            cell = soup.find(text=key)
            if cell:
                return cell.find_parent("td").find_next_sibling("td").get_text().strip()
            else:
                raise KeyError("Could not find cell key %s" % key)
        downloaded = self.store.downloaded_path(doc.basefile)
        fragment = self.store.downloaded_path(doc.basefile, attachment="fragment.html")
        soup = BeautifulSoup(util.readfile(fragment, encoding="utf-8"))
        desc = Describer(doc.meta, doc.uri)
        desc.rdftype(self.rdf_type)
        desc.value(self.ns['prov'].wasGeneratedBy, self.qualified_class_name())
        desc.value(self.ns['dct'].identifier, "ARN %s" % doc.basefile)
        desc.value(self.ns['rpubl'].arendenummer, nextcell("Änr"))
        desc.value(self.ns['rpubl'].avgorandedatum, nextcell("Avgörande"))
        desc.value(self.ns['dct'].subject, nextcell("Avdelning"), lang="sv")
        title = util.normalize_space(soup.table.find_all("tr")[3].get_text())
        title = re.sub("Avgörande \d+-\d+-\d+; \d+-\d+\.?", "", title)
        desc.value(self.ns['dct'].title, title.strip(), lang="sv")
        # dct:issued is required, but we only have rpubl.avgorandedatum
        desc.value(self.ns['dct'].issued,
                desc.getvalue(self.ns['rpubl'].avgorandedatum))
        filetype = os.path.splitext(downloaded)[1]
        if filetype == ".pdf":
            # make sure it really is a PDF
            doctype = self.guess_type(downloaded)
            if doctype != ".pdf":
                util.robust_rename(downloaded,
                                   downloaded.replace(".pdf", doctype))
                downloaded = downloaded.replace(".pdf", doctype)
                filetype = doctype

            self.parse_from_pdf(doc, downloaded, filetype=filetype)

        # for uniformity, we try to treat everything as PDF by
        # converting, even .doc/.docx documents are converted to PDF
        # elif filetype in (".doc", ".docx"):
        # self.parse_from_word(doc, downloaded)
        else:
            self.parse_from_pdf(doc, downloaded, filetype=filetype)
        return True

    def parse_from_pdf(self, doc, filename, filetype=".pdf"):
        def gluecondition(textbox, nextbox, prevbox):
            linespacing = 7
            return (textbox.getfont() == nextbox.getfont() and 
                    textbox.top + textbox.height + linespacing >= nextbox.top)
            
        reader = PDFReader()
        convert_to_pdf = filetype != ".pdf"
        workdir = os.path.dirname(self.store.intermediate_path(doc.basefile))
        if self.config.compress == "bz2":
            keep_xml = "bz2"
        else:
            keep_xml = True
        reader.read(filename, workdir,
                    images=self.config.pdfimages,
                    convert_to_pdf=convert_to_pdf,
                    keep_xml=keep_xml)
        textboxes = reader.textboxes(gluecondition)
        doc.body = Body(textboxes)

        if not hasattr(self, 'ref_parser'):
            self.ref_parser = LegalRef(LegalRef.RATTSFALL, LegalRef.LAGRUM, LegalRef.FORARBETEN)
        citparser = SwedishCitationParser(self.ref_parser)
        citparser.parse_recursive(doc.body)

    def create_external_resources(self, doc):
        pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import codecs
import copy

import six


class TextReader(object):

    """Fancy file-like-class for reading (not writing) text files by line,
    paragraph, page or any other user-defined unit of text, with
    support for peeking ahead and looking backwards. It can read
    files with byte streams using different encodings, but
    converts/handles everything to real strings (unicode in python
    2). Alternatively, it can be initialized from an existing
    string.

    :param filename: The file to read
    :type filename: str
    :param encoding: The encoding used by the file (default ``ascii``)
    :type encoding: str
    :param string: Alternatively, a string used for initialization
    :type string: str
    :param linesep: The line separators used in the file/string
    :type linesep: str
    """

    def __init__(self, filename=None, encoding=None, string=None, linesep=None):
        if not filename and not string:
            raise TypeError("Must specify either filename or string")

        # implementation of file attributes
        self.closed = False
        self.mode = "r+"
        self.name = filename
        self.newlines = None
        self.softspace = 0

        if encoding:
            self.encoding = encoding
        else:
            self.encoding = 'ascii'

        # Other initialization
        if linesep:
            self.linesep = linesep
        else:
            self.linesep = os.linesep

        # can be changed through getiterator, if we want to iterate over anything else but lines
        self.iterfunc = self.readline
        self.iterargs = []
        self.iterkwargs = {}
        self.autostrip = False
        self.autodewrap = False
        self.autodehyphenate = False
        self.expandtabs = True
        if filename:
            self.f = codecs.open(self.name, "r", self.encoding)
            self.data = self.f.read()
            self.f.close()
        else:
            assert(isinstance(string, six.text_type))
            self.data = string

        self.currpos = 0
        self.maxpos = len(self.data)
        self.lastread = ''

    UNIX = '\n'
    """Unix line endings, for use with the ``linesep`` parameter."""
    DOS = '\r\n'
    """Dos/Windows line endings, for use with the ``linesep`` parameter."""
    MAC = '\r'
    """Old-style Mac line endings, for use with the ``linesep`` parameter."""

    def __iter__(self):
        # self.iterfunc = self.readline
        return self

    def __find(self, delimiter, startpos):
        idx = self.data.find(delimiter, startpos)
        if idx == -1:  # not found, read until eof
            res = self.data[startpos:]
            newpos = startpos + len(res)
        else:
            res = self.data[startpos:idx]
            newpos = idx + len(delimiter)
        return (res, newpos)

    def __rfind(self, delimiter, startpos):
        idx = self.data.rfind(delimiter, 0, startpos)
        if idx == -1:  # not found, read until bof
            res = self.data[:startpos]
            newpos = 0
        else:
            res = self.data[idx + len(delimiter):startpos]
            newpos = idx
        return (res, newpos)

    def __process(self, s):
        if self.autostrip:
            s = self.__strip(s)
        if self.autodewrap:
            s = self.__dewrap(s)
        if self.autodehyphenate:
            s = self.__dehyphenate(s)
        if self.expandtabs:
            s = self.__expandtabs(s)
        return s

    def __strip(self, s):
        return s.strip()

    def __dewrap(self, s):
        return s.replace(self.linesep, " ")

    def __dehyphenate(self, s):
        return s  # FIXME: implement

    def __expandtabs(self, s):
        return s.expandtabs(8)

    #----------------------------------------------------------------
    # Added convenience methods

    def eof(self):
        """Returns True iff current seek position is at end of file."""
        return (self.currpos == self.maxpos)

    def bof(self):
        """Returns True iff current seek position is at begining of file."""
        return (self.currpos == 0)

    def cue(self, string):
        """Set seek position at the beginning of *string*, starting at current seek position. Raises IOError if *string* not found."""
        idx = self.data.find(string, self.currpos)
        if idx == -1:
            raise IOError("Could not find %r in file" % string)
        self.currpos = idx

    def cuepast(self, string):
        """Set seek position at the beginning of *string*, starting at current seek position. Raises IOError if *string* not found."""
        self.cue(string)
        self.currpos += len(string)

    def readto(self, string):
        """Read and return all text between current seek potition and *string*. Sets new seek position at the start of *string*.  Raises IOError if *string* not found."""
        idx = self.data.find(string, self.currpos)
        if idx == -1:
            raise IOError("Could not find %r in file" % string)
        res = self.data[self.currpos:idx]
        self.currpos = idx
        return self.__process(res)

    def readparagraph(self):
        """Reads and returns the next paragraph (all text up to
two or more consecutive line separators)."""
        # consume any leading newlines
        while self.peek(len(self.linesep)) == self.linesep:
            self.currpos += len(self.linesep)

        # read actual paragrapgh
        res = self.readchunk(self.linesep * 2)
        # consume any trailing lines
        while self.peek(len(self.linesep)) == self.linesep:
            self.currpos += len(self.linesep)

        # print("readparagraph: %r" % res[:40])
        return res

    def readpage(self):
        """Reads and returns the next page (all text up to next form feed, ``"\\f"``)"""

        return self.readchunk('\f')  # form feed - pdftotext generates
                                    # these to indicate page breaks
                                    # (other ascii oriented formats,
                                    # like the GPL, RFCs and even some
                                    # python source code, uses it as
                                    # well)

    def readchunk(self, delimiter):
        """Reads and returns the next chunk of text up to *delimiter*"""
        (self.lastread, self.currpos) = self.__find(delimiter, self.currpos)
        return self.__process(self.lastread)

    def lastread(self):
        """Returns the last chunk of data that was actually read (i.e. the ``peek*`` and ``prev*`` methods do not affect this)"""
        return self.__process(self.lastread)

    def peek(self, size=0):
        """Works like :meth:`~ferenda.TextReader.read`, but does not affect current seek position."""
        res = self.data[self.currpos:self.currpos + size]
        return self.__process(res)

    def peekline(self, times=1):
        """Works like :meth:`~ferenda.TextReader.readline`, but does not affect current seek position. If *times* is specified, peeks that many lines ahead."""
        return self.peekchunk(self.linesep, times)

    def peekparagraph(self, times=1):
        """Works like :meth:`~ferenda.TextReader.readparagraph`, but does not affect current seek position. If *times* is specified, peeks that many paragraphs ahead."""
        startpos = self.currpos
        # consume any leading newlines
        while self.peek(len(self.linesep)) == self.linesep:
            self.currpos += len(self.linesep)

        # read actual paragrapgh
        res = self.peekchunk(self.linesep * 2, times)

        # print("peekparagraph: %r" % res[:40])
        self.currpos = startpos
        return res

    def peekchunk(self, delimiter, times=1):
        """Works like :meth:`~ferenda.TextReader.readchunk`, but does not affect current seek position. If *times* is specified, peeks that many chunks ahead."""
        oldpos = self.currpos
        for i in range(times):
            (res, newpos) = self.__find(delimiter, oldpos)
            # print "peekchunk: newpos: %s, oldpos: %s" % (newpos,oldpos)
            if newpos == oldpos:
                raise IOError("Peek past end of file")
            else:
                oldpos = newpos
        return self.__process(res)

    def prev(self, size=0):
        """Works like :meth:`~ferenda.TextReader.read`, but reads backwards from current seek position, and does not affect it."""
        res = self.data[self.currpos - size:self.currpos]
        return self.__process(res)

    def prevline(self, times=1):
        """Works like :meth:`~ferenda.TextReader.readline`, but reads backwards from current seek position, and does not affect it. If *times* is specified, reads the line that many times back."""
        return self.prevchunk(self.linesep, times)

    def prevparagraph(self, times=1):
        """Works like :meth:`~ferenda.TextReader.readparagraph`, but reads backwards from current seek position, and does not affect it. If *times* is specified, reads the paragraph that many times back."""
        return self.prevchunk(self.linesep * 2, times)

    def prevchunk(self, delimiter, times=1):
        """Works like :meth:`~ferenda.TextReader.readchunk`, but reads backwards from current seek position, and does not affect it. If *times* is specified, reads the chunk that many times back."""
        oldpos = self.currpos
        for i in range(times):
            (res, newpos) = self.__rfind(delimiter, oldpos)
            if newpos == oldpos:
                raise IOError("Prev (backwards peek) past end of file")
            else:
                oldpos = newpos
        return self.__process(res)

    def getreader(self, callableObj, *args, **kwargs):
        """Enables you to treat the result of any single ``read*``, ``peek*``
        or ``prev*`` methods as a new TextReader. Particularly useful to
        process individual pages in page-oriented documents::

            filereader = TextReader("rfc822.txt")
            firstpagereader = filereader.getreader(filereader.readpage)
            # firstpagereader is now a standalone TextReader that only
            # contains the first page of text from rfc822.txt
            filereader.seek(0) # reset current seek position
            page5reader = filereader.getreader(filereader.peekpage, times=5)
            # page5reader now contains the 5th page of text from rfc822.txt

        """
        res = callableObj(*args, **kwargs)
        clone = copy.copy(self)
        clone.data = res
        clone.currpos = 0
        clone.maxpos = len(clone.data)
        return clone

    def getiterator(self, callableObj, *args, **kwargs):
        """Returns an
        iterator::

            filereader = TextReader("dashed.txt")
            # dashed.txt contains paragraphs separated by "----"
            for para in filereader.getiterator(filereader.readchunk, "----"):
                print(para)
        """
        self.iterfunc = callableObj
        self.iterargs = args
        self.iterkwargs = kwargs
        return self

    #----------------------------------------------------------------
    # Implementation of a file-like interface
    def flush(self):
        """See :py:meth:`io.IOBase.flush`. This is a no-op."""

    def read(self, size=0):
        """See :py:meth:`io.TextIOBase.read`."""
        self.lastread = self.data[self.currpos:self.currpos + size]
        self.currpos += len(self.lastread)
        return self.__process(self.lastread)

    def readline(self, size=None):
        """See :py:meth:`io.TextIOBase.readline`.

        .. note::

           The ``size`` parameter is not supported."""
        # FIXME: the size arg is required for file-like interfaces,
        # but we don't support it
        return self.readchunk(self.linesep)

    def seek(self, offset, whence=0):
        """See :py:meth:`io.TextIOBase.seek`.

        .. note::

           The ``whence`` parameter is not supported."""
        self.currpos = offset

    def tell(self):
        """See :py:meth:`io.TextIOBase.tell`."""
        return self.currpos

    def write(str):
        """See :py:meth:`io.TextIOBase.write`.

        .. note::

           Always raises IOError, as TextReader is a read-only object."""
        return IOError("TextReaders are read-only")

    def writelines(sequence):
        """See :py:meth:`io.IOBase.writelines`.

        .. note::

           Always raises IOError, as TextReader is a read-only object."""
        return IOError("TextReaders are read-only")

    def __next__(self):
        oldpos = self.currpos
        # res = self.__process(self.readline())
        # print "self.iterfunc is %r" % self.iterfunc
        res = self.__process(self.iterfunc(*self.iterargs, **self.iterkwargs))
        if self.currpos == oldpos:
            raise StopIteration
        else:
            return res

    # alias for py2 compat
    next = __next__
    """Backwards-compatibility alias for iterating over a file in python
2. Use :meth:`~ferenda.TextReader.getiterator` to make iteration work over anything other
than lines (eg paragraphs, pages, etc).

    """



# -*- coding: utf-8 -*-
"""Most of these decorators are intended to handle various aspects of
a complete :py:meth:`~ferenda.DocumentRepository.parse`
implementation. Normally you should only use the
:py:func:`~ferenda.decorators.managedparsing` decorator (if you even
override the basic implementation). If you create separate actions
aside from the standards (``download``, ``parse``, ``generate`` et
al), you should also use :py:func:`~ferenda.decorators.action` so that
manage.py will be able to call it.
"""
from __future__ import unicode_literals
from datetime import datetime
import codecs
import functools
import itertools
import os
import time

import six
from rdflib import Graph, URIRef

from ferenda import util
from ferenda import LayeredConfig
from ferenda.errors import DocumentRemovedError, ParseError
from ferenda.elements import serialize

def timed(f):
    """Automatically log a statement of how long the function call takes"""
    @functools.wraps(f)
    def wrapper(self, doc):
        start = time.time()
        ret = f(self, doc)
        # FIXME: We shouldn't log this if we don't actually do any
        # work. The easiest way is to make sure parseifneeded wraps
        # timed, not the other way round.
        self.log.info('%s: OK (%.3f sec)', doc.basefile, time.time() - start)
        return ret
    return wrapper


def recordlastdownload(f):
    """Automatically stores current time in ``self.config.lastdownloaded``
    """
    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        ret = f(self, *args, **kwargs)
        self.config.lastdownload = datetime.now()
        LayeredConfig.write(self.config)
        return ret
    return wrapper


def parseifneeded(f):
    """Makes sure the parse function is only called if needed, i.e. if
    the outfile is nonexistent or older than the infile(s), or if the
    user has specified in the config file or on the command line that
    it should be re-generated."""
    @functools.wraps(f)
    def wrapper(self, doc):
        # note: We hardcode the use of .parseneeded() and the
        # 'parseforce' config option, which means that this decorator
        # can only be used sensibly with the .parse() function.
        force = (self.config.force is True or
                 self.config.parseforce is True)
        if not force and not self.parseneeded(doc.basefile):
            self.log.debug("%s: Skipped", doc.basefile)
            return True  # Signals that everything is OK
        else:
            self.log.debug("%s: Starting", doc.basefile)
            return f(self, doc)
    return wrapper


def render(f):
    """Handles the serialization of the :py:class:`~ferenda.Document`
    object to XHTML+RDFa and RDF/XML files. Must be used in
    conjunction with :py:func:`~ferenda.decorators.makedocument`.

    """
    # NOTE: The actual rendering is two lines of code. The bulk of
    # this function validates that the XHTML+RDFa file that we end up
    # with contains the exact same triples as is present in the doc
    # object (including both the doc.meta Graph and any other Graph
    # that might be present on any doc.body object)
    
    def iterate_graphs(node):
        res = []
        if hasattr(node, 'meta') and node.meta is not None:
            res.append(node.meta)
        for subnode in node:
            if not isinstance(subnode, six.string_types):
                res.extend(iterate_graphs(subnode))
        return res

    @functools.wraps(f)
    def wrapper(self, doc):
        # call the actual function that creates the doc data
        ret = f(self, doc)
        # now render thath doc data as files (JSON, XHTML, RDF/XML)
        if self.config.serializejson == True:
            with self.store.open_serialized(doc.basefile, "wb") as fp:
                r = serialize(doc, format="json") # should be a (unicode) str
                fp.write(r.encode('utf-8'))
            self.log.debug("%s: Created %s" % (doc.basefile, self.store.serialized_path(doc.basefile)))

        updated = self.render_xhtml(doc, self.store.parsed_path(doc.basefile))
        if updated:
            self.log.debug("%s: Created %s" % (doc.basefile, self.store.parsed_path(doc.basefile)))

        # css file + background images + png renderings of text
        self.create_external_resources(doc)

        # Extract all triples on the XHTML/RDFa data to a separate
        # RDF/XML file
        distilled_graph = Graph()
        with codecs.open(self.store.parsed_path(doc.basefile),
                         encoding="utf-8") as fp:  # unicode
            distilled_graph.parse(data=fp.read(), format="rdfa",
                                  publicID=doc.uri)


        # The act of parsing from RDFa binds a lot of namespaces
        # in the graph in an unneccesary manner. Particularly it
        # binds both 'dc' and 'dcterms' to
        # 'http://purl.org/dc/terms/', which makes serialization
        # less than predictable. Blow these prefixes away.
        distilled_graph.bind("dc", URIRef("http://purl.org/dc/elements/1.1/"))
        distilled_graph.bind(
            "dcterms",
            URIRef("http://example.org/this-prefix-should-not-be-used"))

        util.ensure_dir(self.store.distilled_path(doc.basefile))
        with open(self.store.distilled_path(doc.basefile),
                  "wb") as distilled_file:
            # print("============distilled===============")
            # print(distilled_graph.serialize(format="turtle").decode('utf-8'))
            distilled_graph.serialize(distilled_file, format="pretty-xml")
        self.log.debug(
            '%s: %s triples extracted to %s', doc.basefile,
            len(distilled_graph), self.store.distilled_path(doc.basefile))

        # Validate that all required triples are present (we check
        # distilled_graph, but we could just as well check doc.meta)
        for p in self.required_predicates:
            x = distilled_graph.value(URIRef(doc.uri), p)
            if not x:
                self.log.warning("%s: Metadata is missing a %s triple" %
                                 (doc.basefile, distilled_graph.qname(p)))

        # Validate that all triples specified in doc.meta and any
        # .meta property on any body object is present in the
        # XHTML+RDFa file.
        for g in iterate_graphs(doc.body):
            doc.meta += g

        for triple in distilled_graph:
            # len_before = len(doc.meta)
            doc.meta.remove(triple)
            # len_after = len(doc.meta)

        if doc.meta:
            self.log.warning("%s: %d triple(s) from the original metadata was "
                             "not found in the serialized XHTML file:\n%s",
                             doc.basefile, len(doc.meta),
                             doc.meta.serialize(format="nt").decode('utf-8').strip())
        return ret
    return wrapper


def handleerror(f):
    """Make sure any errors in :py:meth:`ferenda.DocumentRepository.parse`
    are handled appropriately and do not stop the parsing of all documents.
    """
    @functools.wraps(f)
    def wrapper(self, doc):
        try:
            return f(self, doc)
        except DocumentRemovedError as e:
            self.log.info(
                "%s: Document has been removed (%s)", doc.basefile, e)
            util.robust_remove(self.parsed_path(doc.basefile))
            return False
        except KeyboardInterrupt:
            raise
        except ParseError as e:
            self.log.error("%s: ParseError %s", doc.basefile, e)
            if (hasattr(self.config, 'fatalexceptions') and
                    self.config.fatalexceptions):
                raise
            else:
                return False
        except:
            self.log.exception("parse of %s failed", doc.basefile)
            if (hasattr(self.config, 'fatalexceptions') and
                    self.config.fatalexceptions):
                raise
            else:
                return False
    return wrapper


def makedocument(f):
    """Changes the signature of the parse method to expect a Document
    object instead of a basefile string, and creates the object."""
    @functools.wraps(f)
    def wrapper(self, basefile):
        doc = self.make_document(basefile)
        return f(self, doc)
    return wrapper


def managedparsing(f):
    """Use all standard decorators for parse() in the correct order
    (:py:func:`~ferenda.decorators.makedocument`, :py:func:`~ferenda.decorators.parseifneeded`, :py:func:`~ferenda.decorators.timed`, :py:func:`~ferenda.decorators.render`)"""
    return makedocument(
        parseifneeded(
            # handleerror( # is this really a good idea?
            timed(
                render(f))))


def action(f):
    """Decorator that marks a class or instance method as runnable by
    :py:func:`ferenda.manager.run`
    """
    f.runnable = True
    return f


def downloadmax(f):
    """Makes any generator respect the ``downloadmax`` config parameter.

    """
    @functools.wraps(f)
    def wrapper(self, params):
        downloadmax = None
        if 'FERENDA_DOWNLOADMAX' in os.environ:
            downloadmax = int(os.environ['FERENDA_DOWNLOADMAX'])
        elif self.config.downloadmax:
            downloadmax = self.config.downloadmax

        if downloadmax:
            self.log.info("Downloading max %d documents" %
                          (downloadmax))
            generator = itertools.islice(f(self, params),
                                         downloadmax)
        else:
            self.log.debug("Downloading all the docs")
            generator = f(self, params)
        for value in generator:
            yield value
    return wrapper


def newstate(state):
    def real_decorator(f):
        setattr(f, 'newstate', state)
        return f
    return real_decorator



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

# the main idea is to just make sure every line of code is run once,
# not to instantiate all eight different
# implementations/configurations and run them all. This will make the
# test code mimick the implementation to some extent, but as the plan
# is to mock all http requests/RDFLib calls (neither of which is
# idempotent), that is sort of unavoidable.

import json, re, os, sqlite3
from tempfile import mkstemp, mkdtemp
import shutil

import pyparsing
from rdflib import Graph, URIRef, RDFS, Literal
import requests.exceptions

from ferenda.compat import patch, Mock, unittest
from ferenda import util, errors
from ferenda.testutil import FerendaTestCase

# SUT
from ferenda import TripleStore

# we could have a switch in canned() that, if set, actually calls
# the request.get or post methods and writes the result to the
# given files.
def canned(*responses):
    returned = []
    def makeresponse(*args, **kwargs):
        if len(returned) > len(responses):
            raise IndexError("Ran out of canned responses after %s calls" % len(returned))
        resp = Mock()
        resp.status_code = responses[len(returned)][0]
        responsefile = responses[len(returned)][1]
        if responsefile:
            responsefile = "test/files/triplestore/" + responsefile
            resp.content = util.readfile(responsefile, "rb")
            resp.text = util.readfile(responsefile)
            if responsefile.endswith(".json"):
                data = json.loads(util.readfile(responsefile))
                resp.json = Mock(return_value=data)
        returned.append(True)
        return resp
    return makeresponse
        
class Main(unittest.TestCase, FerendaTestCase):

    @patch('ferenda.triplestore.util.runcmd')
    def test_curl(self, runcmd_mock):
        # needs to test add_serialized, add_serialized_file, get_serialized
        # and get_serialized_file. We'll patch util.runcmd and make sure that
        # the command line is correct. We should also have util.runcmd return
        # a non-zero return code once.
        # our util.runcmd replacement should, for the get_serialized file,
        # create a suitable temp file

        store = TripleStore.connect("FUSEKI", "", "", curl=True)
        # 1. add_serialized
        runcmd_mock.return_value = (0, "", "")
        store.add_serialized("tripledata", "nt")
        cmdline = runcmd_mock.call_args[0][0] # first ordered argument
        # replace the temporary file name
        cmdline = re.sub('"@[^"]+"', '"@tempfile.nt"', cmdline)
        self.assertEqual('curl -X POST --data-binary "@tempfile.nt" --header "Content-Type:text/plain;charset=UTF-8" "//data?default"', cmdline)
        runcmd_mock.mock_reset()

        # 2. add_serialized_file
        runcmd_mock.return_value = (0, "", "")
        store.add_serialized_file("tempfile.nt", "nt")
        cmdline = runcmd_mock.call_args[0][0] # first ordered argument
        self.assertEqual('curl -X POST --data-binary "@tempfile.nt" --header "Content-Type:text/plain;charset=UTF-8" "//data?default"', cmdline)
        runcmd_mock.mock_reset()

        # 3. get_serialized
        def create_tempfile(*args, **kwargs):
            filename = re.search('-o "([^"]+)"', args[0]).group(1)
            with open(filename, "wb") as fp:
                fp.write("tripledata\n".encode())
            return (0, "", "")
        runcmd_mock.side_effect = create_tempfile
        res = store.get_serialized("nt")
        self.assertEqual(b"tripledata\ntripledata\n", res)
        cmdline = runcmd_mock.call_args[0][0] # first ordered argument
        # replace the temporary file name
        cmdline = re.sub('-o "[^"]+"', '-o "tempfile.nt"', cmdline)
        # FIXME is this really right?
        self.assertEqual('curl -o "tempfile.nt" --header "Accept:text/plain" "//data?graph=urn:x-arq:UnionGraph"', cmdline)
        runcmd_mock.side_effect = None
        runcmd_mock.mock_reset()

        # 4. get_serialized_file
        store.get_serialized_file("triples.xml", "xml")
        cmdline = runcmd_mock.call_args[0][0] # first ordered argument
        self.assertEqual('curl -o "triples.xml" --header "Accept:application/rdf+xml" "//data?default"', cmdline)
        runcmd_mock.mock_reset()

        # 5. handle errors
        with self.assertRaises(errors.TriplestoreError):
            runcmd_mock.return_value = (1, "", "Internal error")
            store.get_serialized_file("triples.nt", "nt")

    def test_fuseki_initialize_triplestore(self):
        store = TripleStore.connect("FUSEKI", "", "")
        store.initialize_repository()

        store = TripleStore.connect("FUSEKI", "http://localhost/", "mydataset")
        store.initialize_repository()
        
    @patch('requests.get', side_effect=canned(("200", "defaultgraph.nt"),
                                             ("200", "namedgraph.nt"),
                                             ("200", "namedgraph.nt"),
                                             ("200", "defaultgraph.ttl"),
                                             ("200", "namedgraph.ttl")))
    def test_fuseki_get_serialized_file(self, mock_get):
        # Test 1: imagine that server has data in the default graph
        # and in one named graph
        rf = util.readfile
        tmp = mkdtemp()
        try:
            store = TripleStore.connect("FUSEKI", "", "")
            # test 1.1: Get everything, assert that the result is a combo
            store.get_serialized_file(tmp+"/out.nt") # no ctx, will result in 2 gets
            self.assertEqual(mock_get.call_count, 2)
            self.assertEqual(rf("test/files/triplestore/combinedgraph.nt"),
                             rf(tmp+"/out.nt"))
            # test 1.2: Get only namedgraph, assert that only that is returned
            store.get_serialized_file(tmp+"/out.nt", context="namedgraph") # 1 get
            self.assertEqual(rf("test/files/triplestore/namedgraph.nt"),
                             rf(tmp+"/out.nt"))
            self.assertEqual(mock_get.call_count, 3)
            # test 1.3: Get everything in a different format
            store.get_serialized_file(tmp+"/out.ttl", format="turtle") # results in 2 gets
            self.assertEqualGraphs("test/files/triplestore/combinedgraph.ttl",
                                  tmp+"/out.ttl")
            self.assertEqual(mock_get.call_count, 5)
        finally:
            shutil.rmtree(tmp)
                
    @patch('requests.get', side_effect=canned(("200", "namedgraph.nt"),))
    def test_fuseki_get_serialized(self, mock_get):
        store = TripleStore.connect("FUSEKI", "", "", curl=False)
        # test 1: a namedgraph (cases with no context are already run by
        # test_fuseki_get_serialized_file)
        want = util.readfile("test/files/triplestore/namedgraph.nt", "rb")
        got = store.get_serialized(context="namedgraph") # results in single get
        self.assertEqual(want, got)

    @patch('requests.delete')
    @patch('requests.post')
    def test_fuseki_clear(self, mock_post, mock_delete):
        store = TripleStore.connect("FUSEKI", "", "")
        store.clear()
        self.assertEqual(mock_delete.call_count, 0)
        self.assertEqual(mock_post.call_count, 1)
        with self.assertRaises(errors.TriplestoreError):
            mock_post.side_effect = requests.exceptions.ConnectionError("Server error")
            got = store.clear()

        with self.assertRaises(errors.TriplestoreError):
            mock_post.side_effect = requests.exceptions.HTTPError("Server error")
            got = store.clear()

        mock_post.side_effect = requests.exceptions.HTTPError("No such graph")
        got = store.clear("namedgraph")


    @patch('requests.get', side_effect=canned(("200", "triplecount-21.xml"),
                                             ("200", "triplecount-18.xml"),
                                             ("200", "triplecount-18.xml")))
    def test_fuseki_triple_count(self, mock_get):
        store = TripleStore.connect("FUSEKI", "", "")
        self.assertEqual(39, store.triple_count())
        self.assertEqual(mock_get.call_count, 2)
        self.assertEqual(18, store.triple_count(context="namedgraph"))
        self.assertEqual(mock_get.call_count, 3)


    @patch('requests.post', side_effect=canned((204, None),
                                               (204, None)))
    def test_fuseki_add_serialized_file(self, mock_post):
        store = TripleStore.connect("FUSEKI", "", "")
        store.add_serialized_file("test/files/triplestore/defaultgraph.ttl",
                                  format="turtle")
        self.assertEqual(mock_post.call_count, 1)

    @patch('requests.get', side_effect=canned(("200", "ping.txt"),))
    def test_sesame_ping(self, mock_get):
        store = TripleStore.connect("SESAME", "", "")
        self.assertEqual("5", store.ping())

    def test_sesame_initialize_triplestore(self):
        store = TripleStore.connect("SESAME", "", "")
        store.initialize_repository()

    @patch('requests.get', side_effect=canned(("200", "combinedgraph.nt"),
                                              ("200", "namedgraph.nt")))
    def test_sesame_get_serialized(self, mock_get):
        store = TripleStore.connect("SESAME", "", "")
        want = util.readfile("test/files/triplestore/combinedgraph.nt", "rb")
        got = store.get_serialized() 
        self.assertEqual(want, got)
        self.assertEqual(mock_get.call_count, 1)

        want = util.readfile("test/files/triplestore/namedgraph.nt", "rb")
        got = store.get_serialized(context="namedgraph") # results in single get
        self.assertEqual(want, got)
        self.assertEqual(mock_get.call_count, 2)

    @patch('requests.post', side_effect=canned((204, None),
                                               (204, None)))
    def test_sesame_add_serialized(self, mock_post):
        store = TripleStore.connect("SESAME", "", "")
        rf = util.readfile
        store.add_serialized(rf("test/files/triplestore/defaultgraph.ttl"),
                             format="turtle")
        self.assertEqual(mock_post.call_count, 1)

        store.add_serialized(rf("test/files/triplestore/namedgraph.nt"),
                             format="nt",
                             context="namedgraph")
        self.assertEqual(mock_post.call_count, 2)

   
    @patch('requests.get', side_effect=canned((200, "select-results.xml"),
                                              (200, "select-results.json"),
                                              (200, "select-results.xml")))
    def test_sesame_select(self, mock_get):
        store = TripleStore.connect("SESAME", "", "")
        rf = util.readfile
        want = rf("test/files/triplestore/select-results.xml")
        got = store.select("the-query")
        self.assertEqual(want, got)
        self.assertEqual(mock_get.call_count, 1)

        want = rf("test/files/triplestore/select-results.json")
        got = store.select("the-query", format="json")
        self.assertEqual(json.loads(want), json.loads(got))
        self.assertEqual(mock_get.call_count, 2)

        # want = json.loads(rf("test/files/triplestore/select-results-python.json"))
        want = json.loads(rf("test/files/triplestore/select-results-python.json"))
        got = store.select("the-query", format="python")
        self.assertEqual(want, got)
        self.assertEqual(mock_get.call_count, 3)

        with self.assertRaises(errors.TriplestoreError):
            mock_get.side_effect = requests.exceptions.HTTPError("Server error")
            got = store.select("the-query", format="python")
            

    
    @patch('requests.get', side_effect=canned((200, "construct-results.xml")))
    def test_sesame_construct(self, mock_get):
        store = TripleStore.connect("SESAME", "", "")
        rf = util.readfile
        want = Graph()
        want.parse(data=rf("test/files/triplestore/construct-results.ttl"),
                   format="turtle")
        got = store.construct("the-query")
        self.assertEqualGraphs(want, got)
        self.assertEqual(mock_get.call_count, 1)

        with self.assertRaises(errors.TriplestoreError):
            mock_get.side_effect = requests.exceptions.HTTPError("Server error")
            got = store.construct("the-query")
        
        
    @patch('requests.get', side_effect=canned(("200", "size-39.txt"),
                                             ("200", "size-18.txt")))
    def test_sesame_triple_count(self, mock_get):
        store = TripleStore.connect("SESAME", "", "")
        self.assertEqual(39, store.triple_count())
        self.assertEqual(mock_get.call_count, 1)
        self.assertEqual(18, store.triple_count(context="namedgraph"))
        self.assertEqual(mock_get.call_count, 2)
        

    @patch('ferenda.triplestore.ConjunctiveGraph')
    def test_sqlite_init(self, mock_graph):
        # create a new db that doesnt exist
        mock_graph.open.return_value = 42
        store = TripleStore.connect("SQLITE", "", "")
        self.assertTrue(mock_graph.return_value.open.called)
        self.assertTrue(mock_graph.return_value.open.call_args[1]['create'])

        # reopen an existing db
        fd, tmpname = mkstemp()
        fp = os.fdopen(fd)
        fp.close()
        store = TripleStore.connect("SQLITE", tmpname, "")
        os.unlink(tmpname)
        self.assertFalse(mock_graph.return_value.open.call_args[1]['create'])

        # make an inmemory db
        store = TripleStore.connect("SQLITE", "", "", inmemory=True)
        self.assertTrue(mock_graph.return_value.quads.called)
        self.assertTrue(mock_graph.return_value.addN.called)

    @patch('ferenda.triplestore.ConjunctiveGraph')
    def test_sqlite_add_serialized(self, mock_graph):
        store = TripleStore.connect("SQLITE", "", "")
        store.add_serialized("tripledata", "nt")
        self.assertTrue(mock_graph.return_value.parse.called)
        self.assertTrue(mock_graph.return_value.commit.called)
        mock_graph.reset_mock()
        
        store.add_serialized("tripledata", "nt", "namedgraph")
        self.assertTrue(mock_graph.return_value.get_context.called)
        self.assertTrue(mock_graph.return_value.get_context.return_value.parse.called)

        store = TripleStore.connect("SQLITE", "", "", inmemory=True)
        with self.assertRaises(errors.TriplestoreError):
            store.add_serialized("tripledata", "nt")

        
    @patch('ferenda.triplestore.ConjunctiveGraph')
    def test_sqlite_add_serialized_file(self, mock_graph):
        store = TripleStore.connect("SQLITE", "", "")
        fd, tmpname = mkstemp()
        fp = os.fdopen(fd, "w")
        fp.write("tripledata")
        fp.close()
        store.add_serialized_file(tmpname, "nt")
        os.unlink(tmpname)

    @patch('ferenda.triplestore.ConjunctiveGraph')
    def test_sqlite_get_serialized(self, mock_graph):
        store = TripleStore.connect("SQLITE", "", "")
        mock_graph.return_value.serialize.return_value = "tripledata"
        self.assertEqual(store.get_serialized(), "tripledata")

    @patch('ferenda.triplestore.ConjunctiveGraph')
    def test_sqlite_triple_count(self, mock_graph):
        store = TripleStore.connect("SQLITE", "", "")
        self.assertEqual(0, store.triple_count())

    @patch('ferenda.triplestore.ConjunctiveGraph')
    def test_sqlite_select(self, mock_graph):
        store = TripleStore.connect("SQLITE", "", "")
        sq = """SELECT ?p FROM <http://example.org/ctx> WHERE {?s ?p ?o . }"""
        res = mock_graph.return_value.get_context.return_value.query.return_value
        want = [{"s": "http://example.org/doc1",
                 "p": "http://www.w3.org/2000/01/rdf-schema#comment",
                 "o": "Hello"}]
        res.bindings = want
        self.assertEqual(want, store.select(sq, format="python"))
        mock_graph.reset_mock()
        store.select(sq, "sparql")
        mock_graph.return_value.get_context.return_value.query.return_value.serialize.assert_called_with(format="xml")
        
        store.select(sq, "json")
        mock_graph.return_value.get_context.return_value.query.return_value.serialize.assert_called_with(format="json")
        
        mock_graph.return_value.get_context.return_value.query.side_effect = pyparsing.ParseException("Syntax error")
        with self.assertRaises(errors.SparqlError):
            store.select(sq)
        
        
    @patch('ferenda.triplestore.ConjunctiveGraph')
    def test_sqlite_construct(self, mock_graph):
        store = TripleStore.connect("SQLITE", "", "")
        sq = """CONSTRUCT ?s ?p ?o WHERE {?o ?p ?s . }"""
        g = Graph()
        g.add((URIRef("http://example.org/doc1"), RDFS.comment, Literal("Hey")))
        g.add((URIRef("http://example.org/doc2"), RDFS.comment, Literal("Ho")))
        res = Mock
        res.graph = g
        mock_graph.return_value.query.return_value = res
        self.assertEqual(g, store.construct(sq))
    
        mock_graph.return_value.query.side_effect = pyparsing.ParseException("Syntax error")
        with self.assertRaises(errors.SparqlError):
            store.construct(sq)

    
    @patch('ferenda.triplestore.ConjunctiveGraph')
    def test_sqlite_clear(self, mock_graph):
        store = TripleStore.connect("SQLITE", "", "")
        g = Graph()
        g.add((URIRef("http://example.org/doc1"), RDFS.comment, Literal("Hey")))
        g.add((URIRef("http://example.org/doc2"), RDFS.comment, Literal("Ho")))
        mock_graph.return_value.get_context.return_value = g
        store.clear("namedgraph")
        self.assertEqual(2, mock_graph.return_value.remove.call_count)
        self.assertEqual(1, mock_graph.return_value.commit.call_count)
        
    @patch('ferenda.triplestore.ConjunctiveGraph')
    def test_sqlite_initialize_triplestore(self, mock_graph):
        store = TripleStore.connect("SQLITE", "", "")
        store.initialize_repository()
        self.assertTrue(mock_graph.return_value.open.call_args[1]['create'])
        

    @patch('ferenda.triplestore.ConjunctiveGraph')
    def test_sqlite_remove_repository(self, mock_graph):
        store = TripleStore.connect("SQLITE", "", "")
        store.remove_repository()
        self.assertTrue(mock_graph.return_value.destroy.called)

    @patch('ferenda.triplestore.ConjunctiveGraph')
    def test_sqlite_close(self, mock_graph):
        # make sure this wierd but harmless sqlite3 exception is
        # caught
        mock_graph.return_value.close.side_effect = sqlite3.ProgrammingError("You made a wrong")
        store = TripleStore.connect("SQLITE", "", "")
        store.close()
        


    @patch('ferenda.triplestore.ConjunctiveGraph')
    def test_sleepycat_init(self, mock_graph):
        store = TripleStore.connect("SLEEPYCAT", "", "")
        
    @patch('ferenda.triplestore.ConjunctiveGraph')
    def test_sleepycat_triple_count(self, mock_graph):
        store = TripleStore.connect("SLEEPYCAT", "", "")
        self.assertEqual(0, store.triple_count())

    def test_invalid_store(self):
        with self.assertRaises(ValueError):
            TripleStore.connect("INVALID", "", "")
            



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import os
import datetime
import ast
import logging
import itertools
import tempfile
from ferenda.compat import OrderedDict
from six.moves import configparser
from six import text_type as str
from six import binary_type as bytes


class LayeredConfig(object):

    """Provides unified access to a nested set of configuration
       parameters. The source of these parameters a config file (using
       .ini-file style syntax), command line parameters, and default
       settings embedded in code. Command line parameters override
       configuration file parameters, which in turn override default
       settings in code (hence **Layered** Config).

       Configuration parameters are accessed as regular object
       attributes, not dict-style key/value pairs.  Configuration
       parameter names should therefore be regular python identifiers
       (i.e. only consist of alphanumerical chars and '_').

       Configuration parameter values can be typed (strings, integers,
       booleans, dates, lists...).  Even though ini-style config files
       and command line parameters are by themselves non-typed, by
       specifying default settings in code, parameters from a config
       file or the commamd line can be typed.

       Configuration parameters can be changed in code. Such changes
       can be written to the configuration file by calling
       :py:meth:`~ferenda.LayeredConfig.write`.

       :param defaults: A dict with configuration keys and values. If
                        any values are dicts, these are turned into
                        nested config objects.
       :type defaults: dict

       :param inifile: The name of a ini-style configuration file. The
                       file should have a top-level section named
                       __root__, whose keys are turned into top-level
                       configuration parameters. Any other sections in
                       this file are turned into nested config
                       objects.
       :param inifile: full path to a .ini-style config file. 
       :type inifile: str
       :param commandline: The contents of sys.argv, or something
                           similar. Any long-style parameters are
                           turned into configuration values, and
                           parameters with hyphens are turned into
                           nested config objects
                           (i.e. ``--module-parameter=foo`` results
                           in self.module.parameter == "foo".
       :type commandline: list
       :param cascade: If an attempt to get a non-existing parameter
                       on a sub (nested) configuration object should
                       attempt to get the parameter on the parent
                       config object.
       :param cascade: If a configuration key is not found,
                       search parent config object.
       :type cascade: bool

       Example::

           >>> defaults = {'parameter': 'foo', 'other': 'default'}
           >>> dir = tempfile.mkdtemp()
           >>> inifile = dir + os.sep + "test.ini"
           >>> with open(inifile, "w") as fp:
           ...     res = fp.write("[__root__]\\nparameter = bar")
           >>> argv = ['--parameter=baz']
           >>> conf = LayeredConfig(defaults, inifile, argv)
           >>> conf.parameter == 'baz'
           True
           >>> conf.other == 'default'
           True
           >>> conf.parameter = 'changed'
           >>> conf.other = 'also changed'
           >>> LayeredConfig.write(conf)
           >>> with open(inifile) as fp:
           ...     res = fp.read()
           >>> res == '[__root__]\\nparameter = changed\\nother = also changed\\n\\n'
           True
           >>> os.unlink(inifile)
           >>> os.rmdir(dir)
    """

    def __init__(self, defaults=None, inifile=None, commandline=None, cascade=False):
        self._cascade = cascade

        self._subsections = OrderedDict()

        self._defaults = OrderedDict()
        self._load_defaults(defaults)

        self._inifile = OrderedDict()
        self._load_inifile(inifile)
        self._inifilename = inifile
        self._inifile_dirty = False

        self._commandline = OrderedDict()
        self._load_commandline(commandline)

        self._parent = None
        self._sectionkey = None

    @staticmethod
    def write(config):
        """Write changed properties to inifile (if provided at initialization)."""
        root = config
        while root._parent:
            root = root._parent
        if root._inifile_dirty:
            with open(root._inifilename, "w") as fp:
                root._configparser.write(fp)

    def __iter__(self):
        l = []
        iterables = [self._commandline.keys(),
                     self._inifile.keys(),
                     self._defaults.keys()]

        if self._cascade and self._parent:
            iterables.append(self._parent)
        
        for k in itertools.chain(*iterables):
            if k not in l:
                l.append(k)
                yield k

    def __getattribute__(self, name):
        if name.startswith("_") or name == "write":
            return object.__getattribute__(self, name)

        if name in self._subsections:
            return self._subsections[name]

        if name in self._commandline:
            return self._commandline[name]
        if self._cascade:
            current = self._parent
            while current:
                if name in current._commandline:
                    return current._commandline[name]
                current = current._parent

        if name in self._inifile:
            return self._inifile[name]
        if self._cascade:
            current = self._parent
            while current:
                if name in current._inifile:
                    return current._inifile[name]
                current = current._parent

        if name in self._defaults:
            if isinstance(self._defaults[name], type):
                return None  # it's typed and therefore exists, but has no value
            else:
                return self._defaults[name]
        if self._cascade:
            current = self._parent
            while current:
                if name in current._defaults:
                    if isinstance(current._defaults[name], type):
                        return None  # it's typed and therefore exists, but has no value
                    else:
                        return current._defaults[name]
                current = current._parent

        raise AttributeError("Configuration key %s doesn't exist" % name)

    def __setattr__(self, name, value):
        # print("__setattribute__ %s to %s" % (name,value))
        if name.startswith("_"):
            object.__setattr__(self, name, value)
            return

        # First make sure that the higher-priority
        # commandline-derived data doesn't shadow the new value.
        if name in self._commandline:
            del self._commandline[name]
        # then update our internal representation
        if name not in self._inifile:
            self._inifile[name] = None
        if value != self._inifile[name]:
            self._inifile[name] = value
            root = self
            while root._parent:
                root = root._parent
            if root._inifilename:
                root._inifile_dirty = True

        # and finally update our associated cfgparser object so that we later
        # can write() out the inifile. This lasts part is a bit complicated as
        # we need to find the root LayeredConfig object where the cfgparser
        # object resides.
        root = self
        sectionkeys = []
        while root._parent:
            # print("root._parent is not None")
            sectionkeys.append(root._sectionkey)
            root = root._parent

        branch = root._configparser
        if branch:
            section = "".join(sectionkeys)  # doesn't really work with more than 1 level
            # print("Setting %s %s %s" % (section,name,str(value)))
            if not section:
                section = "__root__"
            root._configparser.set(section, name, str(value))
        else:
            # If there is no _configparser object, then this
            # LayeredConfig instance was created without one. There is
            # no way to persist configuration values, so we're done
            pass

    def _load_defaults(self, defaults):
        if not defaults:
            return
        for (k, v) in defaults.items():
            if isinstance(v, dict):
                self._subsections[k] = LayeredConfig(
                    defaults=v, cascade=self._cascade)
                self._subsections[k]._sectionkey = k
                self._subsections[k]._parent = self
            else:
                self._defaults[k] = v

    def _load_inifile(self, inifilename):
        if not inifilename:
            self._configparser = None
            return
        if not os.path.exists(inifilename):
            logging.warn("INI file %s does not exist" % inifilename)
            self._configparser = None
            return

        self._configparser = configparser.ConfigParser(dict_type=OrderedDict)
        self._configparser.read(inifilename)

        if self._configparser.has_section('__root__'):
            self._load_inifile_section('__root__')
        for sectionkey in self._configparser.sections():

            # Do we have a LayeredConfig object for sectionkey already?
            if sectionkey not in self._subsections:
                self._subsections[
                    sectionkey] = LayeredConfig(cascade=self._cascade)
                self._subsections[sectionkey]._sectionkey = sectionkey
                self._subsections[sectionkey]._parent = self

            if self._subsections[sectionkey]._configparser is None:
                self._subsections[
                    sectionkey]._configparser = self._configparser

            self._subsections[sectionkey]._load_inifile_section(sectionkey)
        # return cfgparser

    def _type_value(self, key, value):
        """Find appropriate method/class constructor to convert a
           string value to the correct type IF we know the correct
           type."""
        def boolconvert(value):
            # not all bools should be converted, see test_typed_commandline
            if value == "True":
                return True
            elif value == "False":
                return False
            else:
                return value
            
        def listconvert(value):
            # this function is called with both string represenations
            # of entire lists and simple (unquoted) strings. The
            # ast.literal_eval handles the first case, and if the
            # value can't be parsed as a python expression, it is
            # returned verbatim (not wrapped in a list, for reasons)
            try:
                return ast.literal_eval(value)
            except (SyntaxError, ValueError):
                return value

        def datetimeconvert(value):
            try:
                return datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S.%f")
            except ValueError:
                return datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")

        # find the appropriate defaults object. In the case of
        # cascading, could be any parent that has a key key.
        defaults = self._defaults  # base case
        if self._cascade:
            cfg_obj = self
            while cfg_obj is not None:
                if key in cfg_obj._defaults:
                    defaults = cfg_obj._defaults
                    break  # done!
                if hasattr(cfg_obj, '_parent'):
                    cfg_obj = cfg_obj._parent
                else:
                    cfg_obj = None

        if key in defaults:
            if type(defaults[key]) == type:
                # print("Using class for %s" % key)
                t = defaults[key]
            else:
                # print("Using instance for %s" % key)
                t = type(defaults[key])
        else:
            t = str

        if t == bool:
            t = boolconvert
        elif t == list:
            t = listconvert
        elif t == datetime.datetime:
            t = datetimeconvert
        # print("Converting %r to %r" % (value,t(value)))
        return t(value)

    def _load_inifile_section(self, sectionname):
        for (key, value) in self._configparser.items(sectionname):
            self._inifile[key] = self._type_value(key, value)

    # For now: only support long arguments with = separating the parameter and the value, ie
    # "./foo.py --longarg=value" works, "./foo.py --longarg value" or even
    # "./foo.py --longarg = value" doesn't work.
    def _load_commandline(self, commandline):
        if not commandline:
            return
        for arg in commandline:
            if isinstance(arg, bytes):
                arg = arg.decode("utf-8") # FIXME: Find out proper way
                                          # of finding the encoding of
                                          # argv
            if arg.startswith("--"):
                if "=" in arg:
                    (param, value) = arg.split("=", 1)
                else:
                    (param, value) = (arg, True)  # assume bool, not str
                # '--param' => ['param']
                # '--module-param' => ['module','param']
                # Note: Options may not contains hyphens (ie they can't
                # be called "parse-force") since this clashes with hyphen
                # as the sectionkey separator.
                parts = param[2:].split("-")

                self._load_commandline_part(parts, value)

    def _load_commandline_part(self, parts, value):
        if len(parts) == 1:
            key = parts[0]
            if type(value) != bool:  # bools are created implicitly for value-less options
                value = self._type_value(key, value)
            # create a new value, or append to an existing list?
            if key in self._commandline:
                if not isinstance(self._commandline[key], list):
                    self._commandline[key] = [self._commandline[key]]
                self._commandline[key].append(value)
            else:
                self._commandline[key] = value

        else:
            (sectionkey) = parts[0]
            if sectionkey not in self._subsections:
                self._subsections[
                    sectionkey] = LayeredConfig(cascade=self._cascade)
                self._subsections[sectionkey]._sectionkey = sectionkey
                self._subsections[sectionkey]._parent = self
            self._subsections[sectionkey]._load_commandline_part(parts[1:], value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import os
import sys
sys.path.append("..")

try:
    from ferenda.manager import make_wsgi_app

    # FIXME: should we chdir to os.path.dirname(__file__) instead?
    inifile = os.path.join(os.path.dirname(__file__), "ferenda.ini")
    application = make_wsgi_app(inifile=inifile)
except ImportError as e:
    exception_data = str(e)

    def application(environ, start_response):
        msg = """500 Internal Server Error: %s

        sys.path: %r
        os.getcwd(): %s""" % (exception_data, sys.path, os.getcwd())
        msg = msg.encode('ascii')
        start_response("500 Internal Server Error", [
            ("Content-Type", "text/plain"),
            ("Content-Length", str(len(msg)))
        ])
        return iter([msg])
